"""
Concordia plotting routines for (dis)equilibrium and U-Pb datasets

"""

import warnings
import numpy as np

from scipy import integrate
from scipy.interpolate import interp1d

from . import plotting
from . import cfg
from . import misc
from . import useries

from .uxpb import ludwig


exp = np.exp


#=================================
# Concordia plotting routines
#=================================

def plot_concordia(ax, diagram, plot_markers=True, env=False,
                   age_ellipses=False, marker_max=None, marker_ages=(),
                   auto_markers=True, remove_overlaps=True, age_prefix='Ma'):
    """
    Add U-Pb concordia curve to concordia diagram. Raises a warning if
    concordia curve is outside axis limits and returns.
    
    Parameters
    ----------
    marker_max: float, optional
        User specified age marker max (Ma).
    
    """
    assert diagram in ('wc', 'tw')
    # get hard age bounds
    t_bounds = cfg.conc_age_bounds
    if auto_markers and (marker_max is not None):
        if not marker_max > t_bounds[0]:
            raise ValueError('marker_max value must be greater than the lower conc_age_bound value')

    # get t values at axis extremes
    t_limits, nsegs, code = age_limits(ax, diagram, t_bounds=t_bounds)
    if nsegs > 1:
        raise ValueError('cannot have more than 1 concordia segment '
                         'for equilibrium concordia')
    if code != 0:
        return  # outside axis limits

    t_limits = t_limits[0]

    # get equally spaced points
    xy = equi_points(*t_limits, ax.get_xlim(), ax.get_ylim(), diagram)[1:]
    # plot line
    ax.plot(*xy, **cfg.conc_line_kw, label='concordia line')
    # plot envelope
    if env:
        plot_envelope(ax, diagram)
    # plot markers
    if plot_markers:
        if auto_markers and (marker_max is not None) and (marker_max < t_limits[1]):
            t_limits[1] = marker_max
            if marker_max < t_limits[0]:
                warnings.warn('marker_max age is less than auto lower limit - no markers to plot')
                return
        markers = get_age_markers(ax, *t_limits, t_bounds, diagram, auto=auto_markers,
                                  marker_ages=marker_ages, ell=age_ellipses,
                                  age_prefix=age_prefix)
        markers = plot_age_markers(ax, markers)
        # label markers
        if cfg.individualised_labels:
            individualised_labels(ax, markers, diagram,
                                  remove_overlaps=remove_overlaps)
        else:
            labels(ax, markers)


def plot_diseq_concordia(ax, A, init, diagram, sA=None, age_ellipses=False,
                         plot_markers=True, marker_max=None, env=False, env_method='mc',
                         marker_ages=(), auto_markers=True, spaghetti=False, pA=None,
                         remove_overlaps=True, env_trials=1_000, negative_ar=True,
                         age_prefix='Ma', ):
    """
    Add disequilibrium U-Pb concordia curve to concordia plot. Raises a warning if
    concordia curve is outside axis limits and returns.
    
    Parameters
    ----------
    spaghetti : bool 
        plot each simulated line using arbitrary colours
    negative_ar : bool
        if true, allow simulated negative activity ratio values (not yet implemented)

    """
    # get hard age limits depending on which (if any) activity ratios are
    # present values
    if not init[1]:
        t_bounds = cfg.diseq_conc_age_bounds[2]
    elif not init[0]:
        t_bounds = cfg.diseq_conc_age_bounds[1]
    else:
        t_bounds = cfg.diseq_conc_age_bounds[0]

    # get t-limits
    t_limits, nsegs, code = age_limits(ax, diagram, eq=False, A=A, init=init,
                                       t_bounds=t_bounds)

    if code != 0:
        return  # outside axis limits

    for s in range(nsegs):
        # get equally spaced x, y points
        t, x, y = diseq_equi_points(*t_limits[s], ax.get_xlim(), ax.get_ylim(),
                                    A, init, diagram)

        # plot line
        ax.plot(x, y, **cfg.conc_line_kw, label='concordia line')
        # plot envelope
        if env:
            if s > 0:
                warnings.warn('concordia envelope for first segment only will be plotted')
            else:
                if sA is None or all([x == 0 for x in sA]):
                    warnings.warn('cannot plot disequilibrium concordia age ellipses if no '
                              'uncertainity assigned to activity ratios')
                else:
                    assert env_method in ('mc', 'analytical')
                    if env_method == 'mc':
                        # only plot envelope for first segment for now
                        mc_diseq_envelope(ax, t_limits[s], t_bounds, A, sA,
                                  diagram='tw', init=init, trials=env_trials,
                                  spaghetti=spaghetti, negative_ar=negative_ar,
                                  pA=pA)
                    else:
                        analytical_diseq_envelope(ax, t, A, sA, 'tw', init=init)

    # plot markers
    if plot_markers:
        if auto_markers and marker_max is not None and marker_max < t_limits[0][1]:
            t_limits[0][1] = marker_max
            if marker_max < t_limits[0][0]:
                return
        if age_ellipses and (sA is None or all([x == 0 for x in sA])):
            warnings.warn('cannot plot disequilibrium concordia age ellipses if no '
                          'uncertainity assigned to activity ratios')
            return

        # plot markers for first segment...
        markers = get_age_markers(ax, *t_limits[0], t_bounds, diagram, A=A, sA=sA,
                                  init=init, eq=False, ell=age_ellipses, auto=auto_markers,
                                  marker_ages=marker_ages, negative_ar=negative_ar,
                                  age_prefix=age_prefix)
        markers = plot_age_markers(ax, markers)
        # plot markers for second segment...
        if nsegs > 1:
            if auto_markers and (marker_max is not None) and marker_max < t_limits[1][1]:
                t_limits[1][1] = marker_max
                if marker_max < t_limits[1][0]:
                    return
            markers = segment_2_markers(t_limits, markers)
            markers = plot_age_markers(ax, markers)
        # label markers
        if cfg.individualised_labels:
            individualised_labels(ax, markers, diagram, eq=False, A=A, init=init,
                                  remove_overlaps=remove_overlaps)
        else:
            labels(ax, markers)


def segment_2_markers(t_bounds, markers):
    """ Function for adding markers to the second concordia
    segment (uses spacing etc. from first segement).
    """
    t0 = markers['t'][0]
    dt = markers['dt']
    t1, t2 = t_bounds[1]
    t = np.arange(t0, t2 + dt, dt)
    if np.all(markers['add_label']):
        add_label = [True for x in range(len(t))]
    else:
        den = 2 if markers['add_label'] else 1
        add_label = [True if i % den == 0 else False for i, x in
                        enumerate(range(len(t)))]
    markers['t'] = t
    markers['add_label'] = add_label
    return markers


#=================================
# Eq concordia functions
#=================================

def conc_xy(t, diagram):
    """
    Return x, y for given t along (equilibrium) concordia curve.
    """
    assert diagram in ('tw', 'wc')
    if diagram == 'tw':
        x = 1. / (exp(cfg.lam238 * t) - 1.)
        y = (1. / cfg.U) * (exp(cfg.lam235 * t) - 1.) / (exp(cfg.lam238 * t) - 1.)
    elif diagram == 'wc':
        y = exp(cfg.lam238 * t) - 1.
        x = exp(cfg.lam235 * t) - 1.
    return x, y


def conc_age_x(x, diagram):
    """
    Age of point on concordia at given x value.
    """
    assert diagram in ('tw', 'wc')
    if diagram == 'wc':
        t = 1. / cfg.lam235 * np.log(1. + x)
    else:
        t = 1. / cfg.lam238 * np.log(1. + 1. / x)
    return t


def conc_slope(t, diagram):
    """
    Compute tangent to concordia at given t. I.e. dy/dx for given t.
    """
    assert diagram in ('tw', 'wc')
    lam238, lam235, U = cfg.lam238, cfg.lam235, cfg.U
    if diagram == 'wc':
        return lam238 / lam235 * exp((lam238 - lam235) * t)
    elif diagram == 'tw':
        den = (exp(lam238 * t) - 1.)
        dx = -lam238 * exp(lam238 * t) / den ** 2
        dy = lam235 * exp(lam235 * t) / (U * den) - (lam238 *
                exp(lam238 * t) * (exp(lam235 * t) - 1.)) / (U * den ** 2)
        return dy / dx


def conc_age_ellipse(t, diagram):
    """
    Age ellipse params for displaying effects of decay constant
    errors on equilibrium concordia age markers. Requires computing uncertainty
    in x and y for a given t value using first-order error propagation.
    """
    assert diagram in ('tw', 'wc')
    lam238, lam235, s238, s235 = cfg.lam238, cfg.lam235, cfg.s238, cfg.s235
    if diagram == 'wc':
        sx = t * exp(lam235 * t) * s235
        sy = t * exp(lam238 * t) * s238
        cov_xy = 0. * t                         # float or array of zeros
    else:
        x, y = conc_xy(t, diagram)
        sx = - x ** 2 * t * exp(lam238 * t) * s238
        sy = x * t * np.sqrt((exp(lam235 * t) * s235 / cfg.U) ** 2 + (
                (y * exp(lam238 * t) * s238) ** 2))
        cov_xy = x ** 3 * y * (t * exp(lam238 * t) * s238) ** 2
    return sx, sy, cov_xy


def conc_envelope(x, diagram):
    """
    Uncertainty in y for a given x value along the concordia to
    display the effects of decay constant errors on trajectory of the concordia
    curve. Requires computing uncertainty in y for a given x value using
    first-order error propagation.
    """
    assert diagram in ('tw', 'wc')
    lam238, lam235, s238, s235 = cfg.lam238, cfg.lam235, cfg.s238, cfg.s235
    t = conc_age_x(x, diagram)
    if diagram == 'wc':
        sy =  t * exp(lam238 * t) * np.sqrt(s238 ** 2
                + (lam238 / lam235 * s235) ** 2)
    else:
        sy = x * t * exp(lam235 * t) / cfg.U * np.sqrt(s235 ** 2
                + (lam235 / lam238 * s238) ** 2)
    return sy


def velocity(t, xlim, ylim, diagram):
    """
    Estimate dr/dt, which is "velocity" along a diseq concordia curve in
    x-y space. Uses axis coordinates to circumvent scaling issues.
    """
    # TODO: this could be calculated analytically for eq case ??
    h = 1e-08 * t
    xspan = xlim[1] - xlim[0]
    yspan = ylim[1] - ylim[0]
    x2, y2 = conc_xy(t + h, diagram)
    x1, y1 = conc_xy(t - h, diagram)
    v = np.sqrt(((x2 - x1) / xspan) ** 2 + ((y2 - y1) / yspan) ** 2) / (2. * h)
    return v


def equi_points(t1, t2, xlim, ylim, diagram, ngp=500_000, n=500):
    """
    Uses numerical method to obtain age points that are approximately evenly
    spaced in x, y along U-Pb concordia between age limits t1 and t2.
    """
    # TODO: could this be done analytically too?
    assert t2 > t1
    t = np.linspace(t1, t2, ngp)
    # suppress numpy warnings
    with np.errstate(all='ignore'):
        dr = velocity(t, xlim, ylim, diagram)
    # Cumulative integrated area under velocity curve (aka cumulative
    # "distance") at each t_j from t1 to t2:
    cum_r = integrate.cumtrapz(dr, t, initial=0)
    # Divide cumulative area under velocity curve into equal portions.
    rj = np.arange(n + 1) * cum_r[-1] / n
    # Find t_j value at each r_j:
    idx = np.searchsorted(cum_r, rj, side="left")
    idx[-1] = ngp - 1 if idx[-1] >= ngp else idx[-1]
    t = t[idx]
    x, y = conc_xy(t, diagram)
    return t, x, y


#==============================================================
# Diseq concordia functions
#==============================================================

def diseq_xy(t, A, init, diagram):
    """Return x, y for given t along disequilibrium concordia curve.
    """
    assert diagram in ('tw', 'wc')
    DC8 = np.array((cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226))
    DC5 = np.array((cfg.lam235, cfg.lam231))
    BC8 = ludwig.bateman(DC8)
    BC5 = ludwig.bateman(DC5, series='235U')
    if diagram == 'tw':
        x = 1. / ludwig.f(t, A[:-1], DC8, BC8, init=init)
        y = ludwig.g(t, A[-1], DC5, BC5) * x / cfg.U
    elif diagram == 'wc':
        y = ludwig.f(t, A[:-1], DC8, BC8, init=init)
        x = ludwig.g(t, A[-1], DC5, BC5)
    return x, y


def diseq_dxdt(t, A, init, diagram):
    """
    Return x given t along disequilibrium concordia curve.
    Used e.g. to compute dx/dt
    """
    assert diagram == 'tw'
    def conc_x(t):
        x, _ = diseq_xy(t, A, init, diagram)
        return x
    h = abs(t) * np.sqrt(np.finfo(float).eps)
    dxdt = misc.cdiff(t, conc_x, h)
    return dxdt


def diseq_slope(t, A, init, diagram):
    """
    Compute tangent to concordia at given t. I.e. dy/dx for given t.
    """
    h = np.sqrt(np.finfo(float).eps) * t
    x2, y2 = diseq_xy(t + h, A, init, diagram)
    x1, y1 = diseq_xy(t - h, A, init, diagram)
    return (y2 - y1) / (x2 - x1)


def diseq_age_ellipse(t, A, sA, init, diagram, pA=None, trials=1_000,
                     negative_ar=True):
    """
    Plot disequilibrium concordia marker as an "age ellipse" which provides
    a visual representation of uncertainty in x-y for given t value arising from
    uncertainties in activity ratio values.
    """
    if pA is None:
        pA = cfg.rng.normal(A, sA, (trials, 4))
        if not negative_ar:
            warnings.warn('concordia age ellipse plotting routine does not yet account '
                          'for rejected negative activity ratios')
    else:
        assert pA.shape[1] == 4

    flags = np.zeros(trials)
    # check for negative initial [234U/238U] and [230Th/238U] solutions and
    # raise warning if too many found.
    # TODO: in future negative ar solutions may be rejected?
    if not init[0]:
        A48i = useries.ar48i(t, pA[:, 0], cfg.lam238, cfg.lam234)
        flags = np.where(A48i < 0, -1, 0)
    if not init[1]:
        A08i = useries.ar08i(t, pA[:, 0], pA[:, 1], cfg.lam238, cfg.lam234,
                             cfg.lam230, init=init[0])
        flags = np.where((A08i < 0) & (flags == 0), -2, 0)

    if sum(flags != 0) > (0.99 * trials):
        msg = f'{sum(flags != 0)}/{trials} negative activity ratio soln. values in age ellipse t = {t} Ma'
        warnings.warn(msg)

    # Compute ellipse params
    pA = np.transpose(pA)
    xpts, ypts = diseq_xy(t, pA, init, diagram)
    xy = np.vstack((xpts, ypts))

    V = np.cov(xy)
    x, y = np.nanmean(xy, axis=1)
    sx, sy = np.sqrt(V.diagonal())
    cov_xy = V[0, 1]
    r_xy = cov_xy / (sx * sy)

    # reset r_xy for special cases:
    if sA[-1] == 0:
        if any(np.asarray(sA[:2]) != 0):
            r_xy = 1. - 1e-09
            cov_xy = r_xy * sx * sy
        pass
    else:
        if all((x == 0 for x in sA[:2])):
            sx = 0.
            r_xy = 0.
            cov_xy = r_xy * sx * sy

    # cov = np.array([[sx ** 2, r_xy * sx * sy],
    #                 [r_xy * sx * sy, sy ** 2]])

    return x, y, sx, sy, cov_xy


def diseq_equi_points(t1, t2, xlim, ylim, A, init, diagram, ngp=500_000, n=500):
    """
    Return ages that give equally spaced x, y points along disequilbrium
    concordia between upper and lower age limits.

    """
    # TODO: this needs further debugging, sometimes returns t of length n + 1,
    #        also occasionally returns duplicate t values !!!
    assert t2 > t1
    t = np.linspace(t1, t2, ngp)
    # suppress numpy warnings
    with np.errstate(all='ignore'):
        dr = diseq_velocity(t, xlim, ylim, A, init, diagram)
    # Cumulative integrated area under velocity curve (aka cumulative
    # "distance") at each t_j from t1 to t2:
    cum_r = integrate.cumtrapz(dr, t, initial=0)
    # Divide cumulative area under velocity curve into equal portions.
    rj = np.arange(n + 1) * cum_r[-1] / n
    # Find t_j value at each r_j:
    idx = np.searchsorted(cum_r, rj, side="left")
    idx[-1] = ngp - 1 if idx[-1] >= ngp else idx[-1]
    t = t[idx]
    x, y = diseq_xy(t, A, init, diagram)
    return t, x, y


def diseq_velocity(t, xlim, ylim, A, init, diagram):
    """
    Estimate dr/dt, which is "velocity" along a diseq concordia curve in
    x-y space. Uses axis coordinates to circumvent scaling issues.
    """
    h = 1e-08 * t
    xspan = xlim[1] - xlim[0]
    yspan = ylim[1] - ylim[0]
    x2, y2 = diseq_xy(t + h, A, init, diagram)
    x1, y1 = diseq_xy(t - h, A, init, diagram)
    v = np.sqrt(((x2 - x1) / xspan) ** 2 + ((y2 - y1) / yspan) ** 2) / (2. * h)
    return v


#=================================
# Concordia age bound functions
#=================================

def age_limits(ax, diagram, eq=True, A=None, init=(True, True),
               t_bounds=cfg.conc_age_bounds, max_tries=3):
    """
    Find concordia age limits at axis window boundaries. Will return age bound
    if it is inside the axis window.

    """
    if not eq:
        assert A is not None

    code = 0
    tmin, tmax = t_bounds           # maximum and miniumum plotted age
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()

    # Get starting points:
    # Generate evenly spaced points in time and check which are in axis window.
    npts = 100_000
    for i in range(max_tries):
        tv = np.linspace(tmin, tmax, npts)
        # suppress numpy warnings:
        with np.errstate(all='ignore'):
            if eq:
                x, y = conc_xy(tv, diagram)
            else:
                x, y = diseq_xy(tv, A, init, diagram)

        # check if points are inside plot window:
        inside = np.logical_and(np.logical_and(xmin < x, x < xmax),
                                np.logical_and(ymin < y, y < ymax))

        if np.sum(inside) < 10:
            if np.sum(inside) == 0 and i == max_tries - 1:
                warnings.warn('concordia curve appears to lie outside plot window')
                code = 1
                return None, 0, code
            else:
                npts *= 10  # increase number of points and try again
        else:
            break

    limits0, nsegs = t_limits_from_cp(tv, inside, eq=eq, tmin=tmin, tmax=tmax)

    limits = []     # both sets of limits (if 2 segments)

    # Refine age limits by starting at points in box and incrementally
    # moving out until the boundary is encountered:
    # for l in limits0:
    for l0 in limits0:

        l0 = np.sort(l0)
        l = np.zeros(2)

        # Lower age limit
        if l0[0] == t_bounds[0]:    # lower bound inside window - no need to refine
            l[0] = l0[0]
        else:
            l[0], code = refine_t_limit(l0, t_bounds, (xmin, xmax), (ymin, ymax),
                            eq=eq, increasing_t=False, diagram=diagram, ell=False,
                            ax=ax, A=A, init=init)
            if code != 0:
                raise ValueError('failed to find concordia envelope lower age limit')

        # Upper age limit
        if l0[1] == t_bounds[1]:  # upper bound inside window - no need to refine
            l[1] = l0[1]
        else:
            l[1], code = refine_t_limit(l0, t_bounds, (xmin, xmax), (ymin, ymax),
                            eq=eq, increasing_t=True, diagram=diagram, ell=False,
                            A=A, init=init, ax=ax)
            if code != 0:
                raise ValueError('failed to find concordia envelope upper age limit')

        # reset limit if outside bounds
        # TODO: shouldn't need this but just in case?
        if l[0] < t_bounds[0]:
            l[0] = t_bounds[0]
        if l[1] > t_bounds[1]:
            l[1] = t_bounds[1]

        limits.append(l)

    return limits, nsegs, code


def t_limits_from_cp(t, inside, tmin, tmax, eq=True):
    """
    Get estimated concordia age bounds from arbitrary age points within
    the plot window.

    """
    # Get indices of inside / outside change points for different cases:
    idx = np.where(np.diff(inside) != 0)[0]
    limits0 = []
    nsegs = 1

    if eq and len(idx) > 2:
        raise RuntimeError('cannot have more than two change points for '
                           'equilibrium concordia')

    if len(idx) == 0:
        limits0 = [[tmin, tmax]]

    elif len(idx) == 1:  # either tmin or tmax in box
        if inside[0]:
            limits0 = [[tmin, t[idx[0]]]]
        else:
            limits0 = [[t[idx[0] + 1], tmax]]

    elif len(idx) == 2:  # neither tmin nor tmax in box
        limits0 = [[t[idx[1]], t[idx[0] + 1]]]

    else:   # 2 segments
        nsegs = 2
        if len(idx) == 3:  # 3 points inside
            pass

        if len(idx) == 4:  # 4 points inside
            # deal with first segment:
            if idx[1] - idx[0] == 1:    # one point inside
                limits0.append([t[idx[0] + 1], t[idx[0] + 1]])
            else:
                limits0.append([t[idx[0] + 1], t[idx[1]]])

            # deal with second segment:
            if idx[3] - idx[2] == 1:    # one point inside
                limits0.append([t[idx[2] + 1], t[idx[2] + 1]])
            else:
                limits0.append([t[idx[2] + 1], t[idx[3]]])

    return limits0, nsegs


def refine_t_limit(limits0, t_bounds, xlim, ylim, increasing_t,
                    diagram='tw', eq=True, A=None, sA=None, init=(True, True), ell=False,
                    trials=1_000, ax=None):
    """
    Refine concordia age limits so they lie just outside axis window. Will
    return age bound if it is inside the axis window.

    Parameters
    ----------
    tmin : float
        lowder t-bound
    tmax : float
        upper t-bound
    """
    if ell:
        denom = 100.
    else:
        denom = 1000.

    xspread = xlim[1] - xlim[0]
    xtol = xspread / denom
    yspread = ylim[1] - ylim[0]
    ytol = yspread / denom

    t0, t1 = limits0

    if eq:
        x0, y0 = conc_xy(t0, diagram)
        x1, y1 = conc_xy(t1, diagram)
        tpts, xpts, ypts = equi_points(t0, t1, xlim, ylim, diagram, ngp=20_000,
                                       n=20)
    else:
        x0, y0 = diseq_xy(t0, A, init, diagram)
        x1, y1 = diseq_xy(t1, A, init, diagram)
        tpts, xpts, ypts = diseq_equi_points(t0, t1, (x0, x1), (y0, y1), A,
                                 init, diagram, ngp=100_000, n=20)

    # --- testing ----
    # ax.plot(xpts, ypts, 'ro', ms=2)
    # --------

    if ell:
        pA = cfg.rng.normal(A, sA, (trials, 4))

    if increasing_t:
        dt = np.diff(tpts[-2:])[0]  # spacing between last two equi age points
        t_start = t1
    else:
        dt = -np.diff(tpts[:2])[0]  # spacing between first two equi age points
        t_start = t0

    for i in range(100):
        if ell:
            t, code = walk_ellipse_outward(t_start, dt, t_bounds, xlim, ylim, diagram,
                        xtol, ytol, trials, A=A, init=init, pA=pA,
                        ax=ax)
        else:
            t, code = walk_outward(t_start, dt, t_bounds, xlim, ylim, diagram, xtol,
                        ytol, eq=eq, A=A, init=init)
        if code == 0:
            return t, code


        # if walk_outward did not meet tol, reduce and try again
        t_start = t  # reset t0
        dt /= 2.

    # TODO: throw warning, but return t anyway?
    msg = 'age limits did not meet convergence criteria, but may be good enough?'
    warnings.warn(msg)

    return t0, 0


def walk_ellipse_outward(t0, dt, t_bounds, xlim, ylim, diagram, xtol,
                    ytol, trials, A=None, init=None, pA=None, ttol=1e-03,
                    max_steps=100, p=0.999, ax=None):
    """
    Walk outward from t0 until an axis boundary is found (i.e. ellipse is
    pretty much outside axis window, where 'pretty-muchness' is determined
    by p).
    """

    assert diagram == 'tw'
    xmin, xmax = xlim
    ymin, ymax = ylim
    x0, y0 = diseq_xy(t0, A, init, 'tw')    # centre point
    pA = np.transpose(pA)
    step = 0
    while step < max_steps:
        t1 = t0 + dt
        x1, y1 = diseq_xy(t1, A, init, 'tw')    # centre point
        xpts, ypts = diseq_xy(t1, pA, init, 'tw')   # simulated x-y
        # print(f'ell step {"+" if dt > 0 else "-"}: {step}')

        # TODO: deal with negative_ar?

        if dt > 0 and t1 > t_bounds[1]: # if bounds exceeded while still in box..
            return t_bounds[1], 0
        elif dt < 0 and t0 < t_bounds[0]:
            return t_bounds[0], 0
        elif (np.sum(xpts < xmin) / trials > p) or \
                (np.sum(xpts > xmax) / trials > p) or \
                (np.sum(ypts > ymax) / trials > p) or \
                (np.sum(ypts < ymin) / trials > p):  # outside box
            # check tolerances are met - i.e. check estimate is accurate enough
            if (abs(x1 - x0) < xtol and abs(y1 - y0) < ytol):
                return t1, 0
            else:
                return t0, 1

        # keep walking
        t0, x0, y0 = t1, x1, y1
        step += 1

    code = 1
    return t0, code


def walk_outward(t0, dt, t_bounds, xlim, ylim, diagram, xtol, ytol, eq=True,
                 A=None, init=(True, True), ttol=1e-03, max_steps=1000,
                 ell=False, pA=None):
    """
    Walk outward from t0 until an axis boundary is found.
    """
    # x0, y0 = conc_xy(t0, diagram)
    if eq:
        x0, y0 = conc_xy(t0, diagram)
    else:
        x0, y0 = diseq_xy(t0, A, init, diagram)
    step = 0
    while step < max_steps:
        t1 = t0 + dt
        if eq:
            x1, y1 = conc_xy(t1, diagram)
        else:
            x1, y1 = diseq_xy(t1, A, init, diagram)

        if dt > 0 and t1 > t_bounds[1]: # if bounds exceeded while still in box..
            return t_bounds[1], 0
        elif dt < 0 and t0 < t_bounds[0]:
            return t_bounds[0], 0
        elif (x1 > xlim[1] or x1 < xlim[0]) or (y1 > ylim[1] or y1 < ylim[0]):  # outside box
            if (abs(x1 - x0) < xtol and abs(y1 - y0) < ytol or
                    abs(t1 - t0) / t0 < ttol):
                return t1, 0
            else:
                return t0, 1

        # keep walking
        t0, x0, y0 = t1, x1, y1
        step += 1
    code = 1
    return np.nan, code


#====================
# Concordia markers
#====================

def estimate_marker_spacing(tspan):
    """
    Get initial estimate of good concordia marker spacing.

    """
    dt = 10 ** misc.get_exponent(tspan) / 8
    while abs(tspan / dt) > 12:
        dt *= 2
    return misc.round_down(dt, 8)


def get_age_markers(ax, t1, t2, t_bounds, diagram, eq=True, ell=False, A=None,
                    sA=None, init=None, marker_ages=(), age_prefix='Ma',
                    auto=True, n_segments=1, negative_ar=True):
    """
    Add concordia markers to concordia line and generate label text (but
    does not actually add marker labels).

    Parameters
    ----------
    t1 : float
        lower concordia age in plot,
    t2 : float
        upper concordia age
    ell : bool
        plot age ellipses

    """
    assert t2 > t1
    assert age_prefix in ('ka', 'Ma')
    age_unit = 1. if age_prefix == 'Ma' else 1e-3
    t_limits = (t1, t2)
    dt = None

    if not auto:  # manual age markers

        if len(marker_ages) == 0:
            raise ValueError('ages cannot be empty if auto set to False')
        for i, x in enumerate(marker_ages):
            try:
                fx = float(x) * age_unit
            except ValueError:
                raise ValueError(f'could not convert marker age {x} to a'
                                 f'number')
            else:
                if not cfg.conc_age_bounds[0] < fx < cfg.conc_age_bounds[1]:
                    msg = f'marker age {x} {age_prefix} outside ' \
                          f'conc_age_bounds value set in config'
                    warnings.warn(msg)

        t_sorted = np.sort(np.array(marker_ages, dtype=np.double))
        t = t_sorted * age_unit
        n_inside = len(t[np.logical_and(t1 < t, t2 > t)])
        # which markers to label:
        if n_inside > cfg.every_second_threshold:
            add_label = [True if i % 2 == 0 else False for i, t in enumerate(t)]
        else:
            add_label = [True for t in t]

    else:  # find auto marker locations

        dt = age_marker_spacing(ax, t1, t2, diagram, A=A, init=init, eq=eq)

        # Get marker age points:
        t_start = misc.round_down(np.floor(t1 / dt) * dt, 5)
        t = np.arange(t_start, t2 + dt, dt)

        # Check if there are too many diseq age ellipses in plot window and reduce
        # dt if necessary:
        if ell and not eq:
            t0, dt0 = t, dt
            t, t_start, dt, code = age_ellipse_marker_spacing(ax, t, dt, A, sA, init, t_limits)

            if code != 0:
                warnings.warn('age ellipse spacing routine failed')
                t = t0  # go back to normal marker spacing
                dt = dt0

        # Reset to 0 if sufficiently close in order to avoid labelling problems.
        t = [0. if abs(x) < 1e-9 else x for x in t]
        t = np.array([round(x, 10) for x in t])         # round ages to get around f.p. issues

        # If labelling every second, check which label to start with. Preference
        # starting on label with less significant digits, then preference starting
        # on label ending in 1, and finally preference starting on even number.
        n = len(t)
        if ell:
            n -= 1          # on average, at least one marker will be fully outside?
        start_idx = 0
        step = 1
        if n > cfg.every_second_threshold:
            step = 2
            t0 = misc.round_down(float(t[0]), 8)
            t1 = misc.round_down(float(t[1]), 8)
            s0 = str(t0).rstrip("0")  # first marker in sequence
            s1 = str(t1).rstrip("0")  # second marker in sequence

            if s0 == '':
                # if initial t is 0
                start_idx = 1
            elif s0[-1] == '.':
                start_idx = 0
            elif s1[-1] == '.':
                start_idx = 1
            elif len(s0) > len(s1):
                start_idx = 1
            elif float(s0[-1]) % 2 > 0 and float(s1[-1]) % 2 == 0:
                start_idx = 1

        # Add extra markers to ends, in case they are partly displayed in plot
        # window - but this should not affect label starting age.
        if not ell:
            t = np.arange(t_start - dt, t2 + 2 * dt, dt)
            start_idx = 1 if start_idx == 0 else 0

        t = np.array([round(x, 10) for x in t])         # round ages again
        t = t[(t >= t_bounds[0]) & (t <= t_bounds[1])]  # double check bounds
        num_t = len(t)                                  # new number of markers

        # list of bools indicating which markers to add a label to:
        if n > cfg.every_second_threshold:
            add_label = [True if (i - start_idx) % step == 0
                         else False for i in range(num_t)]
        else:
            add_label = [True] * len(t)

    age_markers = {'diagram': diagram, 'eq': eq, 'A': A, 'sA': sA,
                   'init': init, 'ell': ell, 't': t, 'dt': dt,
                   'add_label': add_label, 'negative_ar': negative_ar,
                   'age_prefix': age_prefix}

    return age_markers


def age_ellipse_limits(ax, t0, dt, t_bounds, A, sA, init, p=0.998,
            trials=1_000, diagram='tw', maxiter=40, negative_ar=True):
    """
    Find concordia age ellipse limits near axis window boundaries. Will return
    age bound if it is inside the axis window.

    Parameters
    ----------
    t0 : array-like
        Guess at limits (points must be inside plot window).
    dt : float
        appropriate step size

    """
    assert diagram == 'tw'
    tmin, tmax = t0
    assert tmax >= tmin
    tspan = tmax - tmin
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    t1, t2 = np.nan, np.nan
    code = 0

    pA = np.transpose(cfg.rng.normal(A, sA, (trials, 4)))

    # Left limit:
    # move outward until all endpoints are outside axis window
    t2 = tmax + dt
    i = 1
    while i <= maxiter:
        i += 1
        xpts, ypts = diseq_xy(t2, pA, init, 'tw')
        if (np.sum(xpts < xmin) / trials > p) or \
            (np.sum(ypts > ymax) / trials > p) or \
            (np.sum(ypts < ymin) / trials > p):
                break
        t2 += dt
        if i == maxiter:
            code = 1
            return code, t1, t2
        if t2 > t_bounds[1]:
            t2 = t_bounds[1]
            break

    # Right limit:
    t1 = tmin
    i = 1
    while i <= maxiter:
        i += 1
        xpts, ypts = diseq_xy(t1, pA, init, 'tw')
        if (np.sum(xpts > xmax) / trials > p) or \
            (np.sum(ypts > ymax) / trials > p) or \
            (np.sum(ypts < ymin) / trials > p):
            break
        t1 -= dt
        if t1 < t_bounds[0]:
            t1 = t_bounds[0]
            break
        if i == maxiter:
            code = 1
            return code, t1, t2

    return code, t1, t2


def age_marker_spacing(ax, t1, t2, diagram, A=None, init=None, eq=True,
                       max_markers=12):
    """
    Estimate reasonable concordia age marker spacing given upper and lower
    age marker limits.

    """
    tspan = t2 - t1
    t_ratio = t2 / t1
    # First estimate of spacing between markers, dt.
    dt = estimate_marker_spacing(tspan)
    if t_ratio > 5 and t2 > 1:
        dt = estimate_marker_spacing(tspan / 4)
    # Increase spacing if too many markers.
    while abs(tspan / dt) > max_markers:
        dt *= 2

    # Check dt after calculaing the fraction of x, y axis spanned by
    # concordia and refine if necessary...
    if eq:
        x_tmin, y_tmin = conc_xy(t1, diagram)
        x_tmax, y_tmax = conc_xy(t2, diagram)
    else:
        x_tmin, y_tmin = diseq_xy(t1, A, init, diagram)
        x_tmax, y_tmax = diseq_xy(t2, A, init, diagram)

    x_frac = abs((x_tmin - x_tmax) / (ax.get_xlim()[1] - ax.get_xlim()[0]))
    y_frac = abs((y_tmin - y_tmax) / (ax.get_ylim()[1] - ax.get_ylim()[0]))

    # Refine dt using some rules of thumb.
    k = 0.6
    for i in range(1, 5):
        k *= 0.5
        if x_frac < k and y_frac < k:
            dt *= 2

    return dt


def age_ellipse_marker_spacing(ax, t, dt, A, sA, init, t_bounds, diagram='tw',
                              maxiter=30, trials=1_000):
    """
    Starting with regular concordia age marker spacing and limits, check if
    these are also suitable for disequilibrium age ellipses.
    """
    assert diagram == 'tw'

    idx = [int(np.floor((len(t) - 1)/2)), int(np.ceil((len(t) - 1)/2))]
    t0 = [t[idx[0]], t[idx[1]]]
    code, t1, t2 = age_ellipse_limits(ax, t0, dt, t_bounds, A, sA, init,
                                      p=0.998, trials=1_000)

    if code != 0:
        return np.nan, np.nan, np.nan, code

    t1 = round(t1, 10)
    t2 = round(t2, 10)

    dt = age_marker_spacing(ax, t1, t2, diagram, A=A, init=init, eq=False,
                            max_markers=8)

    # Get marker age points:
    t_start = misc.round_down(np.floor(t1 / dt) * dt, 5)
    t = np.arange(t_start, t2 + dt, dt)

    return t, t_start, dt, code


def plot_age_markers(ax, markers, p=0.95):
    """
    Add age markers to plot.

    Parameters
    ----------
    markers : dict
        age marker properties, typically returned from calling get_age_markers

    """
    # unpack markers dict
    diagram = markers['diagram']
    eq = markers['eq']
    age_prefix = markers['age_prefix']

    A = markers['A']
    sA = markers['sA']
    init = markers['init']
    negative_ar = markers['negative_ar']

    t = markers['t']
    ell = markers['ell']
    add_label = markers['add_label']

    n = len(t)

    # Plot markers / ellipses.
    if eq:
        x, y = conc_xy(t, diagram)
    else:
        x, y = diseq_xy(t, A, init, diagram)

    if ell:    # plot age markers as ellipses
        ell_obj = []
        bbox = []
        xm = np.empty(n)    # use mean simulated x, y in case rejecting negative_ar
        ym = np.empty(n)
        sx = np.empty(n)
        sy = np.empty(n)
        cov_xy = np.empty(n)

        for i, age in enumerate(t):
            if eq:
                xm[i], ym[i] = conc_xy(age, diagram)
                sx[i], sy[i], cov_xy[i] = conc_age_ellipse(age, diagram)
            else:
                xm[i], ym[i], sx[i], sy[i], cov_xy[i] = diseq_age_ellipse(
                    age, A, sA, init, diagram, negative_ar=negative_ar
                )

        for i in range(n):
            V = np.diag((sx[i], sy[i])) ** 2
            V[0, 1] = V[1, 0] = cov_xy[i]
            # --- choose either this ----
            # ellipse = plotting.confidence_ellipse2(ax, x[i], y[i], V,
            #                     n_std=2.0, **cfg.conc_ellipse_kw)
            # --- or this ----
            r_xy = cov_xy[i] / (sx[i] * sy[i])
            ellipse = plotting.confidence_ellipse(ax, xm[i], sx[i], ym[i], sy[i],
                                 r_xy, p=p, mpl_label=f'age ellipse, {t[i]:.6f} Ma',
                                 ellipse_kw=cfg.conc_age_ellipse_kw,
                                 outline_alpha=False)
            # ----------------
            ell_obj.append(ellipse)

    else:   # plot age markers
        ax.plot(x, y, label='concordia marker', **cfg.conc_markers_kw)

    # Generate marker label text.
    age_unit = 1e-3 if age_prefix == 'ka' else 1.
    if sum(add_label) > 0:
        t_rounded = np.array([float(misc.round_down(age / age_unit, 8))
                              for age in t])
        n_dec = np.max([misc.num_dec_places(x) for x in t_rounded[add_label]
                        if add_label])

        if cfg.prefix_in_label:
            label_format = '{{:,.{}f}} {{}}'.format(n_dec)
            text = [label_format.format(x, age_prefix) for x in t_rounded]

        else:
            label_format = '{{:,.{}f}}'.format(n_dec)
            text = [label_format.format(x) for x in t_rounded]

        markers['text'] = text

    markers['x'] = x
    markers['y'] = y
    markers['add_label'] = add_label
    markers['age_ellipses'] = ell

    if ell:
        markers['bbox'] = bbox
        markers['ell_obj'] = ell_obj
    
    return markers


#=====================
# Concordia envelope
#=====================

def plot_envelope(ax, diagram, npts=100):
    """
    Plot concordia uncertainty envelope which displays effect of decay constant
    errors.
    """
    xx = np.linspace(*ax.get_xlim(), num=100, endpoint=True)
    t = conc_age_x(xx, diagram)
    x, y = conc_xy(t, diagram)
    dy = 1.96 * conc_envelope(xx, diagram)
    ax.fill_between(xx, y + dy, y - dy, label='concordia envelope',
                    **cfg.conc_env_kw)


def mc_diseq_envelope(ax, t_limits, t_bounds, A, sA, diagram='tw',
                init=(True, True), trials=1_000, limit_trials = 1_000,
                spaghetti=False, maxiter=50, negative_ar=True, pA=None):
    """
    Plot uncertainty envelope about disequilibrium concordia based on Monte
    Carlo simulation. This displays uncertainty in trajectory of concordia
    arising from uncertainty in activity ratio values.

    Parameters
    ----------
    t_limits : array-like
        minimum and maximum concordia ages for given axis window
    t_bounds :
        minimum and maximum age bounds for full concordia curve

    """
    assert diagram == 'tw'
    tmin, tmax = t_limits      # actual age bounds for 'middle' curve only
    assert tmax > tmin
    tspan = tmax - tmin
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()

    points = 1000
    interp_points = 250

    # Concordia limits - use as initial guess at envelope t limits
    l0 = np.sort(t_limits)

    # Refine age limits by starting at points in box and incrementally
    # moving out until the boundary is encountered:
    # for l in limits0:
    l = np.zeros(2)

    # Lower age limit
    if l0[0] == t_bounds[0]:    # lower bound inside window - no need to refine
        l[0] = l0[0]
    else:
        l[0], code = refine_t_limit(l0, t_bounds, (xmin, xmax), (ymin, ymax),
                        eq=False, increasing_t=False, diagram=diagram,
                            A=A, sA=sA, init=init, ell=True, ax=ax)
        if code != 0:
            raise ValueError('failed to find concordia envelope lower age limit')

    # Upper age limit
    if l0[1] == t_bounds[1]:  # lower bound inside window - no need to refine
        l[1] = l0[1]
    else:
        l[1], code = refine_t_limit(l0, t_bounds, (xmin, xmax), (ymin, ymax),
                        eq=False, increasing_t=True, diagram=diagram,
                            A=A, sA=sA, init=init, ell=True, ax=ax)
        if code != 0:
            raise ValueError('failed to find concordia envelope upper age limit')

    # reset limit if outside bounds
    # TODO: shouldn't need this?
    if l[0] < t_bounds[0]:
        l[0] = t_bounds[0]
    if l[1] > t_bounds[1]:
        l[1] = t_bounds[1]

    t0, t1 = l
    t, x, y = diseq_equi_points(t0, t1, ax.get_xlim(), ax.get_ylim(), A, init,
                                diagram, n=points)
    t = t[:points]  # sometimes diseq_equi_points returns vector of length n + 1?

    # Remove duplicate t values
    # TODO: find a proper solution to this!
    t, ind = np.unique(t, return_index=True)
    x = x[ind]
    y = y[ind]
    points = len(t)
    # ----

    # reset xmin and xmax
    xmin = np.min(x)
    xmax = np.max(x)

    # Perturb activity ratios
    if pA is None:
        pA = cfg.rng.normal(A, sA, (trials, 4))
        if not negative_ar:
            warnings.warn('concordia envelope plotting does not yet account for '
                          'rejected negative activity ratio trials')
    else:
        assert pA.shape[1] == 4

    # check if dx/dt changes sign, if so, there are multiple y for x, and
    # therefore t needs to be truncated to plot envelope
    dxdt = diseq_dxdt(t, A, init, diagram)
    ind = np.where(np.diff(np.sign(dxdt)) != 0)[0]
    if ind.shape[0] != 0:

        if ind.shape[0] > 1:    # if multiple dx/dt changes, probably a numerical issue computing deriv.
            warnings.warn(f'multiple dx/dt sign changes in concordia - no truncation')

        else:
            ind = np.min(ind)
            t = t[:ind]

            # refine change of sign x-limit....
            markers_dict = dict(
                diagram='tw', eq=False, A=A, sA=sA, init=init,
                negative_ar=True, t=np.array([t[-1]]), ell=True,
                add_label=np.array([False]), age_prefix='Ma')
            plot_age_markers(ax, markers_dict, p=0.99)
            # get ellipse bounds...
            ell_obj = ax.patches[-1]
            ax.get_figure().canvas.draw()
            ell_bbox = ax.transData.inverted().transform(ell_obj.get_extents())
            xmin = np.max(ell_bbox[:, 0])
            ell_obj.remove()
            warnings.warn(f'concordia envelope truncated at t = {t[-1]} Ma because dx/dt changes sign')

            # Get evenly spaced x,y points over truncated t
            t, x, y = diseq_equi_points(t0, t[-1], ax.get_xlim(), ax.get_ylim(), A,
                                        init, diagram, n=points)
            t = t[:points]  # sometimes diseq_equi_points returns vector of length n + 1?
            # Remove duplicate t values
            # TODO: find a proper solution to this!
            t, ind = np.unique(t, return_index=True)
            x = x[ind]
            y = y[ind]
            # ----

            points = len(t)  # sometimes diseq_equi_points returns vector of length n + 1?

    # pre-allocate arrays to store interpolated points
    xv = np.linspace(xmin, xmax, interp_points)
    yv = np.zeros((trials, interp_points))

    # pre-allocate arrays to store simulated curves
    xc = np.zeros((trials, points))
    yc = np.zeros((trials, points))

    flags = np.zeros(points)
    for i in range(trials):
        # check for negative inital [234U/238U] and [230Th/238U] solutions and
        # reject these simulated pA curves - these really mess up envelope!
        if not init[0]:
            if any(useries.ar48i(t, pA[i, 0], cfg.lam238, cfg.lam234) < 0):
                flags[i] = -1
        if not init[1]:
            if any(useries.ar08i(t, *pA[i, :2], cfg.lam238, cfg.lam234,
                                 cfg.lam230, init=init[0]) < 0):
                if flags[i] != -1:
                    flags[i] = -2

        xc[i, :], yc[i, :] = diseq_xy(t, pA[i, :], init, 'tw')
        if spaghetti:
            ax.plot(xc[i, :], yc[i, :], ms=0, lw=0.50)
        else:
            # Sample at common x spacing
            f = interp1d(xc[i, :], yc[i, :], kind='cubic', bounds_error=False,
                         fill_value='extrapolate')
            yv[i, :] = f(xv)

    if sum(flags != 0) != 0:
        msg = f'{sum(flags != 0)} / {points} negative activity ratio soln. values in MC concordia env.'
        warnings.warn(msg)
        warnings.warn('MC concordia envelope may be unreliable for given activity ratio values over this age range')
        # yv = yv[flags == 0]

    # Estimate confidence interval from quantiles:
    y_low, y_hi = np.quantile(yv, (0.025, 0.975), axis=0)
    ax.plot(xv, y_low, **cfg.conc_env_line_kw, label='concordia envelope line')
    ax.plot(xv, y_hi, **cfg.conc_env_line_kw, label='concordia envelope line')
    ax.fill_between(xv, y_low, y2=y_hi, **cfg.conc_env_kw,
                    label='concordia envelope')


def analytical_diseq_envelope(ax, t, A, sA, diagram, init=(True, True)):
    """
    !!! Experimental function !!!

    Plot uncertainty envelope about disequilibrium concordia based on analytical
    error propagation. This displays uncertainty in trajectory of concordia
    arising from uncertainty in activity ratio(s).

    Does not yet permit present [230Th/238U] values and requires further
    overall testing.

    Needs further checking / testing / debugging.
    """
    assert diagram == 'tw'
    if not init[1]:
        # issues to resolve when using present [230Th/238U] values.
        raise ValueError('analytical errors not yet implemented for present '
                         '[230Th/238U] values')

    l8, l4, l0, l6 = cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226
    l5, l1 = cfg.lam235, cfg.lam231
    DC8 = (l8, l4, l0, l6)
    DC5 = (l5, l1)
    BC8 = ludwig.bateman(DC8)
    BC5 = ludwig.bateman(DC5, series='235U')

    F = ludwig.f(t, [A[0], A[1], A[2]], DC8, BC8, init=init)
    dF = ludwig.dfdt(t, A[:-1], init=init, comp=False)

    G = ludwig.g(t, A[-1], DC5, BC5)
    dG = ludwig.dgdt(t, A[-1], comp=False)

    dydt = dydt_func(t, A, F, G, dF, dG, init)
    dtdA = dtdA_func(t, A, init)
    pypA = dydA_func(t, A, F, G, init)  # partial derivative

    # Full derivatives.
    dydA48 = dydt * dtdA[0] + pypA[0]
    dydA08 = dydt * dtdA[1] + pypA[1]
    dydA68 = dydt * dtdA[2] + pypA[2]
    dydA15 = dydt * dtdA[3] + pypA[3]

    # Uncertainty in y at given t values.
    dy = np.sqrt( (sA[0] * dydA48) **2 + (sA[1] * dydA08) **2
                  + (sA[2] * dydA68) **2 + (sA[3] * dydA15) **2)

    # Estimate confidence interval from quantiles:
    x, y = diseq_xy(t, A, init, diagram)
    y_low = y - 2. * dy
    y_hi = y + 2. * dy

    ax.plot(x, y_low, **cfg.conc_line_kw)
    ax.plot(x, y_hi, **cfg.conc_line_kw)
    ax.fill_between(x, y_low, y2=y_hi, **cfg.conc_env_kw)


def dydt_func(t, A, F, G, dF, dG, init):
    """
    Partial derivative of y wrt t for Tera-Wasserburg concordia curve.
    """
    dydt = (dG * F - G * dF) / (cfg.U * F ** 2)
    return dydt


def dtdA_func(t, A, init):
    """
    Partial derivative of t wrt A for U-Pb equations.
    """
    l8, l4, l0, l6 = cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226
    l5, l1 = cfg.lam235, cfg.lam231
    c1, c2, c3, c4, h1, h2, h3, p1, p2 = ludwig.bateman((l8, l4, l0, l6))
    d1, d2 = ludwig.bateman((l5, l1), series='235U')

    DC8 = (l8, l4, l0, l6)
    DC5 = (l5, l1)
    BC8 = (c1, c2, c3, c4, h1, h2, h3, p1, p2)
    BC5 = (d1, d2)

    F1, F2, F3, F4 = ludwig.f_comp(t, A[:-1], DC8, BC8, init=init)
    F = np.sum((F1, F2, F3, F4))
    dF1, dF2, dF3, dF4 = ludwig.dfdt(t, A[:-1], init=init, comp=True)

    G1, G2 = ludwig.g_comp(t, A[-1], DC5, BC5)
    G = np.sum((G1, G2))
    dG1, dG2 = ludwig.dgdt(t, A[-1], comp=True)

    if init[0]:
        f2 = l8/l4 * (h1*exp((l8-l4)*t) + h2*exp((l8-l0)*t) + h3*exp((l8-l6)*t)
                      + exp(l8*t))
        df2 = l8/l4 * (h1*(l8-l4)*exp((l8-l4)*t) + h2*(l8-l0)*exp((l8-l0)*t)
                        + h3*(l8-l6)*exp((l8-l6)*t) + l8*exp(l8*t))
        dw2 = 0.
    else:
        f2 = l8/l4 * (h1*exp(l8*t) + h2*exp((l8-l0+l4)*t)
                      + h3*exp((l8-l6+l4)*t) + exp((l8+l4)*t))
        df2 = l8/l4 * (h1*l8*exp(l8*t) + h2*(l8-l0+l4)*exp((l8-l0+l4)*t)
                + h3*(l8-l6+l4)*exp((l8-l6+l4)*t) + (l8+l4)*exp((l8+l4)*t))
        dw2 = l4 * exp(-l4*t)

    if init[1]:
        f3 = l8/l0 * (p1*exp((l8-l0)*t) + p2*exp((l8-l6)*t) + exp(l8*t))
        df3 = l8/l0 * (p1*(l8-l0)*exp((l8-l0)*t) + p2*(l8-l6)*exp((l8-l6)*t)
                       + l8*exp(l8*t))
        dw3 = 0.
    else:
        if init[0]: # present A08, init ar48
            f3 = l8/l0 * (p1*exp(l8*t) + p2*exp((l8-l6+l0)*t) + exp((l8+l0)*t))
            df3 = l8/l0 * (p1*l8*exp(l8*t) + p2*(l8-l6+l0)*exp((l8-l6+l0)*t)
                           + (l8+l0)*exp((l8+l0)*t))
            dw3 = (A[0]-1)*(l0/(l0-l4))*(-l4*exp(-l4*t) + l0*exp(-l0*t)) \
                    + l0*exp(-l0*t)
        else:       # present A08, init ar48
            msg = 'present [230Th/238U] values not yet permitted'
            raise ValueError(msg)

    f4 = l8/l6 * (exp(l8*t) - exp((l8-l6)*t))
    df4 = l8/l6 * (l8*exp(l8*t) - (l8-l6)*exp((l8-l6)*t))
    g2 = l5/l1 * (exp(l5*t) - exp((l5-l1)*t))
    dg2 = l5/l1 * (l5*exp(l5*t) - (l5-l1)*exp((l5-l1)*t))

    # Compute derivatives
    dA48dt = (-(dF1+dF3+dF4) *f2 - (F-F1-F3-F4) *df2) / f2**2 + dw2
    dA08dt = (-(dF1+dF2+dF4) *f3 - (F-F1-F2-F4) *df3) / f3**2 + dw3
    dA68dt = (-(dF1+dF2+dF3) *f4 - (F-F1-F2-F3) *df4) / f4**2
    dA15dt = (-dG1*g2 - (G-G1)*dg2) / g2**2

    dtdA = [1./dA48dt, 1./dA08dt, 1./dA68dt, 1./dA15dt]

    return dtdA


def dydA_func(t, A, F, G, init):
    """
    Partial derivative of y wrt A for Tera-Wasserburg concordia curve.
    """
    l8, l4, l0, l6 = cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226
    l5, l1 = cfg.lam235, cfg.lam231
    c1, c2, c3, c4, h1, h2, h3, p1, p2 = ludwig.bateman((l8, l4, l0, l6))

    if init[0]:
        f2 = l8/l4 * (h1 * exp((l8-l4)*t) + h2 * exp((l8-l0)*t)
                + h3 * exp((l8-l6)*t) + exp(l8*t))
    else:
        f2 = l8/l4 * exp(l4*t) * (h1 * exp((l8-l4)*t) + h2 * exp((l8-l0)*t)
                + h3 * exp((l8-l6)*t) + exp(l8*t))
    if init[1]:
        f3 = l8/l0 * (p1*exp((l8-l0)*t) + p2*exp((l8-l6)*t) + exp(l8*t))
    else:
        msg = 'present [230Th/238U] values not yet permitted'
        raise ValueError(msg)

    f4 = l8/l6 * (exp(l8*t) - exp((l8-l6)*t))
    g2 = l5/l1 * (exp(l5*t) - exp((l5-l1)*t))

    denom = cfg.U * F ** 2
    dydA48 = -G * f2 / denom
    dydA08 = -G * f3 / denom
    dydA68 = -G * f4 / denom
    dydA15 = g2 / (cfg.U*F)

    return [dydA48, dydA08, dydA68, dydA15]


#====================
# Concordia labels
#====================

def labels(ax, markers):
    """
    Add labels to concordia age markers. Uses the same offset and rotation
    for each marker.

    Parameters
    ----------
    markers : dict
        age marker properties

    """
    # Mask out values for markers that will not be labelled.
    add_label = np.array(markers['add_label'])
    x = np.array(markers['x'])[add_label]
    y = np.array(markers['y'])[add_label]
    txt = np.array(markers['text'])[add_label]
    n = sum(add_label)

    ann = []
    for i in range(n):
        an = ax.annotate(txt[i], (x[i], y[i]), **cfg.conc_text_kw,
                         label='concorida label')
        ann.append(an)
    markers['label_annotations'] = ann
    return markers


def individualised_labels(ax, markers_dict, diagram, eq=True, A=None,
                  init=None, remove_overlaps=True):
    """
    Plot concordia age labels using individualised position and rotation.

    Notes
    -----
    This routine doesn't currently work well for disequilibrium concordia curves
    that curve back around.
    """
    assert diagram in ('tw', 'wc')
    fig = ax.get_figure()
    ell = markers_dict['age_ellipses']

    # Calculate some useful axes properties.
    xmin, xmax = ax.get_xlim()
    xspread = xmax - xmin
    ymin, ymax = ax.get_ylim()
    yspread = ymax - ymin

    # Get axis window extents in display points. Then calculate a "scale factor"
    # for converting slopes from data coordinates to display coordinates.
    ax_bbox = ax.get_window_extent()
    aspect_ratio = ax_bbox.height / ax_bbox.width
    scale_factor = yspread / xspread / aspect_ratio

    # Only pass in text properties to annotate(). We do not want to pass
    # in rotation or position properties for indidividualised labels.
    # TODO: consider if alignment properties could still be passed in?
    allowed_keys = ['alpha', 'backgroundcolor', 'color', 'c',
                    'fontfamily', 'family', 'fontsize', 'size',
                    'stretch', 'fontweight', 'weight', 'zorder',
                    'annotation_clip', 'clip_on']
    text_kw = {}
    for k, v in cfg.conc_text_kw.items():
        if k in allowed_keys:
            text_kw[k] = v

    # Create labels
    label_annotations = []

    # Mask out values for markers_dict that will not be labelled.
    add_label = np.array(markers_dict['add_label'])
    x = np.array(markers_dict['x'])[add_label]
    y = np.array(markers_dict['y'])[add_label]
    t = np.array(markers_dict['t'])[add_label]
    txt = np.array(markers_dict['text'])[add_label]
    n = sum(add_label)
    if ell:
        ell_obj = [b for (a, b) in zip(add_label, markers_dict['ell_obj']) if a]

    # Get annotation bbox properties for dummy annotation. Used to
    # calculate annotation box height (parallel to y-axis) in display
    # coordinates.
    an = ax.annotate(txt[0], (x[0], y[0]), va='center', ha='center',
             rotation=0, **text_kw)
    # fig.canvas.draw()
    anbbox = an.get_window_extent(renderer=fig.canvas.get_renderer())
    h = anbbox.height       # display coordinates
    an.remove()

    # TODO: alignment kwargs could be added back in here?

    text_outline = None
    # ------- testing only ------
    # text_outline = dict(ec='red', lw=0.5, pad=0.0, fc='none')
    # ----------------------------------

    # add each marker
    for i in range(n):
        # Add annotation to figure at arbitraty location (on top of marker),
        # so we can get it's bounding box dimensions.
        an = ax.annotate(txt[i], (x[i], y[i]), va='center', ha='center',
                         label='concordia label', rotation=0, bbox=text_outline,
                         **text_kw)
        # anbbox = an.get_window_extent()
        # fig.canvas.draw()
        an_bbox = an.get_window_extent(renderer=fig.canvas.get_renderer())
        w = an_bbox.width
        x_disp, y_disp = ax.transData.transform((x[i], y[i]))

        # ---testing---
        # ax.plot(x[i], y[i], 'bo', ms=2, zorder=100)
        # ----

        # Get concordia slope (i.e. tangent) at t.
        if eq:
            slope = conc_slope(t[i], diagram)
        else:
            slope = diseq_slope(t[i], A, init, diagram)

        # angle of concordia at marker location in display coords:
        # angle = np.arctan(slope / scale_factor) * 180 / np.pi
        angle = np.arctan2(slope, 1.)
        angle_disp = np.arctan2(slope / scale_factor, 1.)

        # ------ testing only - comment out -------
        #plot line parallel to concordia slope
        # b = slope
        # a = y[i] - b * x[i]
        # ax.plot((xmin, xmax), (a + b * xmin, a + b * xmax)), 'b--', lw=0.5, zorder=100)
        # ---------------------------

        # theta is the angle from positive x axis to offset the text box (for
        # normal markers, this will be orthoganol to the slope of the
        # concordia):
        orth_angle = angle + np.pi / 2.
        orth_angle_disp = angle_disp + np.pi / 2.

        # Find slope and intercept of a line running orthogonal (on scaled plot)
        # to the concordia curve at x, y:
        # b = np.tan(theta * np.pi / 180)
        orth_slope = np.tan(orth_angle)           # true orthogonal line (data coords)
        b_disp = orth_slope * scale_factor   # apparent orthogonol line (display coords)
        b = orth_slope * scale_factor ** 2   # apparent orthogonal line (data coords)

        # x,y in display coords
        # x_disp, y_disp = ax.transData.transform((x[i], y[i]))
        # a_disp = y_disp - b_disp * x_disp

        # ------ testing only - comment out-------
        # plot line orthogonal to slope
        # a = y[i] - b * x[i]
        # xx = np.linspace(*ax.get_xlim())
        # yy = a + b * xx
        # ax.plot(xx, yy, 'm--', lw=0.5, zorder=100)
        # ---------------------------

        if cfg.rotate_conc_labels:
            if cfg.perpendicular_rotation:
                if diagram == 'tw':
                    an.set_rotation(orth_angle_disp * 180 / np.pi)
                elif diagram == 'wc':
                    an.set_rotation(orth_angle_disp * 180 / np.pi)
                d = w / 2.

            else:   # rotate parallel
                if diagram == 'tw':
                    an.set_rotation(angle_disp * 180 / np.pi)
                elif diagram == 'wc':
                     an.set_rotation(angle_disp * 180 / np.pi)
                d = h / 2.
        else:
            # Find intersection points of this orthogonal slope line and text
            # bbox:  # need to call this again after rotation
            a_disp = y_disp - b_disp * x_disp
            ints = plotting.box_line_intersection(an_bbox.xmin, an_bbox.xmax,
                        an_bbox.ymin, an_bbox.ymax, a_disp, b_disp)
            # offset text box by this extra amount (data coords)
            d = np.sqrt(np.sum((ints[1] - ints[0]) ** 2)) / 2.


        # Compute offsets
        f = cfg.offset_factor
        # d is the mimumum offset, f is extra user-defined offset (relative
        #   label height and thus textsize)

        # dx_disp = (d + f * h) * -np.abs(np.cos(orth_angle_disp))
        # dy_disp = (d + f * h) * -np.sign(orth_slope) * np.sin(orth_angle_disp)
        dx_disp = -np.sqrt((d + f * h) ** 2 / (b_disp ** 2  + 1.))
        if diagram == 'tw' and b_disp < 0:
            dx_disp *= -1.
        dy_disp = b_disp * dx_disp



        # If age ellipses, add extra offset label along ellipse axis:
        if ell:
            # ---- testing -----
            # plot bbox around ellipse
            # ellbb = ax.transData.inverted().transform(ell_obj[i].get_extents())
            # rec = Rectangle((np.min(ellbb[:, 0]), np.min(ellbb[:, 1])),
            #                 ell_obj[i].width, ell_obj[i].height,
            #                 fc='none', ec='red', lw=0.5, zorder=100)
            # ax.add_patch(rec)
            # ------------------

            # Get slope and y-int of line projected through ellipse major axis.

            # ---- old appraoch -----
            # WARNING !! This doesn't always work as expected !!
            # e_angle = ell_obj[i].get_angle() * np.pi / 180
            # eb = np.tan(e_angle)
            # eb_disp = eb / scale_factor
            # -------

            # get ellipse bbox from path:
            path_xy  = ax.transData.inverted().transform(ell_obj[i].get_verts())
            path_xy = ell_obj[i].get_verts()
            ind_xmin = np.argmin(path_xy[:, 0])
            ind_xmax = np.argmax(path_xy[:, 0])
            ind_ymin = np.argmin(path_xy[:, 1])
            ind_ymax = np.argmax(path_xy[:, 1])

            xy_xmin = path_xy[ind_xmin, :]
            xy_xmax = path_xy[ind_xmax, :]
            xy_ymin = path_xy[ind_ymin, :]
            xy_ymax = path_xy[ind_ymax, :]

            # --- tesing ----
            # for xy in (xy_xmin, xy_xmax, xy_ymin, xy_ymax):
            #     xy = ax.transData.inverted().transform(xy)
            #     ax.plot(*xy, 'mo', ms=4, zorder=120)
            # -------

            # offset along line parralell / orthogonal to ellipse extreme values
            if diagram == 'tw':
                if np.isclose(xy_ymax[0],  xy_ymin[0]):
                    # eb_disp = 0.
                    # no correlation - offset along conc. slope:
                    eb_disp = b_disp
                else:
                    eb_disp = (xy_ymax[1] - xy_ymin[1]) / (xy_ymax[0] - xy_ymin[0])
            else:
                if np.isclose(xy_xmax[1], xy_xmin[1]):
                    # eb_disp = np.inf
                    # no correlation - offset along conc. slope:
                    eb_disp = b_disp
                else:
                    eb_disp = (xy_xmax[1] - xy_xmin[1]) / (xy_xmax[0] - xy_xmin[0])

            eb = eb_disp * scale_factor
            ea = y[i] - eb * x[i]
            ea_disp = y_disp - eb_disp * x_disp

            # ---- testing -----
            # ax.plot((xmin, xmax), (ea + eb * xmin, ea + eb * xmax),
            #         lw=0.5, c='red', ls='-')
            # xmin_disp = ax.get_window_extent().xmin
            # ymin_disp = ax.get_window_extent().ymin
            # xmax_disp = ax.get_window_extent().xmax
            # ymax_disp = ax.get_window_extent().ymax
            # y1 = ea_disp + eb_disp * xmin_disp
            # y2 = ea_disp + eb_disp * xmax_disp
            # x1, y1 = ax.transData.inverted().transform((xmin_disp, y1))
            # x2, y2 = ax.transData.inverted().transform((xmax_disp, y2))
            # ax.plot((x1, x2), (y1, y2), lw=0.5, c='blue', ls='-')
            # -------

            # find intercept points b/w conc slope line and bbox defined
            # by x, y min and max points on ellipse (calling get_window_extents
            # doesn't seem to work properly all the time?)
            ints = plotting.box_line_intersection(xy_xmin[0], xy_xmax[0],
                                xy_ymin[1], xy_ymax[1], ea_disp, eb_disp)

            if ints.size == 0:
                warnings.warn(f'plotting age ellipse t = {t[i]} Ma failed')
                np.array(markers_dict['add_label'])[i] = False
                an.remove()
                continue

            # ----- testing ----
            # ax.plot(*ax.transData.inverted().transform(ints[0, :]), 'bo', ms=2, zorder=100)
            # ax.plot(*ax.transData.inverted().transform(ints[1, :]), 'bo', ms=2, zorder=100)
            # dd = ints[0, :]  - np.array((x_disp, y_disp))
            # dx = dd[0]
            # dy = dd[1]
            # dist = np.sqrt(dx ** 2 + dy ** 2)
            # -----

            # offset text box by this extra amount (data coords)
            d2 = np.sqrt(np.sum((ints[1] - ints[0]) ** 2)) / 2.

            dx_disp2 = -np.sqrt(d2 ** 2 / (eb_disp ** 2 + 1.))
            dy_disp2 = eb_disp * dx_disp2

            # ---testing---
            # x2, y2 = ax.transData.inverted().transform((x_disp + dx_disp2, y_disp + dy_disp2))
            # ax.plot(x2, y2, 'go', ms=2, zorder=100)
            # ----

            dx_disp += dx_disp2
            dy_disp += dy_disp2

        # Transform x, y from display coords back to data coords and re-set
        # annotation position.
        # xy = ax.transData.inverted().transform((x_disp, y_disp))
        x_dat, y_dat = ax.transData.inverted().transform((x_disp + dx_disp,
                                                          y_disp + dy_disp))
        an.set_position((x_dat, y_dat))


        label_annotations.append(an)

    if remove_overlaps:
        remove_overlapping_labels(label_annotations)

    markers_dict['label_annotations'] = label_annotations
    return markers_dict


def remove_overlapping_labels(an):
    """
    Naive routine for removing overlapping concordia labels older than first
    overlap point.

    an : array-like
        list of concordia age marker / ellipse label annotations
    """
    n = len(an)
    overlap = np.full(n, False)

    # Check if annotation bbox fully (?) overlaps axis bbox,
    # and flag if not.
    for i in range(n - 1):
        b1 = an[i].get_window_extent()  # in display coords
        b2 = an[i + 1].get_window_extent()
        if b1.overlaps(b2):
            overlap[i + 1] = True
        else:
            overlap[i + 1] = False
    n_overlap = np.sum(overlap == True)

    # If overlaps found:
    # 1. If only youngest label overlaps, remove this and keep all others.
    # 2. If n > 5 and only youngest 2 overlap, remove these 2 and keep others.
    # 3. Otherwise remove from oldest down to first overlap.
    if n_overlap > 0:
        if n > 2 and n_overlap == 1 and overlap[0]:
            an[0].remove()
        elif n > 5 and n_overlap == 2 and overlap[0] and overlap[1]:
            an[0].remove()
            an[1].remove()
        else:
            for j in reversed(range(np.argwhere(overlap).min(), n)):
                an[j].remove()



"""
Transform data points, regression fits, etc.

References
----------
.. [Ludwig2000]
    Ludwig, K.R., 2000. Decay constant errors in U–Pb concordia-intercept ages.
    Chemical Geology 166, 315–318.
    https://doi.org/10.1016/S0009-2541(99)00219-3

"""

import numpy as np

from . import cfg


def dp_errors(dp, in_error_type, append_corr=True, row_wise=True, dim=2,
              tol=1e-10):
    """
    Convert data point errors to 1 sigma absolute.

    Parameters
    ----------
    in_error_type : str
        Error type and sigma level, use: {abs/per/rel}{1s/2s}
    append_corr : bool, optional
        Assume error correlations are zero by appending a column of zeros to the end
        of n x 4 data point array.
    row_wise : bool, optional
        True if each row is a data point and each column is a variable. False if
        vice versa.
    dim : int
        1 for single variable data, 2 for diagram data.

    """
    assert in_error_type in ('abs1s', 'abs2s', 'per1s', 'per2s', 'rel1s', 'rel2s')
    dp = np.atleast_2d(dp)
    if dp.ndim > 2:
        raise ValueError('dp must a 1 or 2 dimensional array')

    # TODO: test this
    nan_idx = np.argwhere(np.isnan(dp))
    if nan_idx.size > 0:
        raise ValueError("input data contains empty cells or non-allowed "
                         "values in row(s): %s" % [x // dp.ndim for x in nan_idx])

    # convert to one row per variable
    if not row_wise:
        dp = np.transpose(dp)

    npts, nvar = dp.shape  # number of vars and data points

    a = 0.01 if 'per' in in_error_type else 1.
    b = 0.5 if '2' in in_error_type else 1.

    if nvar not in (2, 4, 5):
        raise ValueError('unexptected number of columns in input data')
    if dim == 1 and nvar != 2:
        raise ValueError('incompatible number of dimensions in input data')
    if dim == 2 and nvar not in (4, 5):
        raise ValueError('incompatible number of dimensions in input data')

    x = dp[:, 0]

    # Convert errors to 1 sigma absolute
    c = x if ('per' in in_error_type or 'rel' in in_error_type) else 1.
    sx = dp[:, 1] * a * b * c

    if nvar == 2:
        return np.array((x, sx))

    if nvar in (4, 5):
        y = dp[:, 2]
        c = y if ('per' in in_error_type or 'rel' in in_error_type) else 1.
        sy = dp[:, 3] * a * b * c

    if nvar == 4 and append_corr:
        cor_xy = np.full(npts, 0.)
    elif nvar == 5:
        if dp.shape[1] == 5:
            # TODO: values within some tolerance of +/- 1 should be reset to min/max values
            if np.any(np.abs(dp[:, -1]) > 1.):
                raise ValueError('data point correlation coefficients must '
                                 'be > -1 and < 1')
        cor_xy = dp[:, 4]

    return np.array((x, sx, y, sy, cor_xy))


def transform_fit(fit, transform_to='wc'):
    """
    Transform regression theta and covtheta from Tera-Wasserburg
    coordinates to Wetheril (conventional discordia) coordinates or vice versa
    in one go.
    """
    theta = transform_theta(fit['theta'], transform_to=transform_to)
    covtheta = transform_covtheta(fit['theta'], fit['covtheta'],
                   transform_to=transform_to)
    return theta, covtheta


def transform_theta(theta, transform_to='wc'):
    """
    Transform theta from Tera-Wasserburg (tw) coordinates to Wetheril
    (conventional concordia, wc) coordinates or vice versa.
    WC int / slope are denoted A / B.
    T-W int / slope are denoted a / b.
    """
    assert transform_to in ('wc', 'tw')
    if transform_to == 'wc':
        a, b = theta
        A = -b / a
        B = 1. / (cfg.U * a)
        return np.array((A, B))
    elif transform_to == 'tw':
        A, B = theta
        a = 1. / (cfg.U * B)
        b = - A / (cfg.U * B)
        return np.array((a, b))


def transform_covtheta(theta0, covtheta0, transform_to='wc'):
    """
    Transform regression fit covtheta from Tera-Wasserburg coordinates to
    Wetheril (conventional discordia) coordinates or vice versa.
    See, e.g., Ludwig (2000), Chemical Geology, 166, 315–318
    """
    assert transform_to in ('tw', 'wc')
    a, b = theta0
    A, B = transform_theta(theta0, transform_to=transform_to)
    if transform_to == 'wc':
        jac = np.array([[b / a ** 2, -1. / a], [1. / (cfg.U * a ** 2), 0. ]])
    else:
        raise ValueError('not yet coded')
    covtheta = jac @ covtheta0 @ jac.T
    #TODO: there is an error here causing incorrect sign for covariance!!!
    covtheta[1, 0] *= -1.
    covtheta[0, 1] *= -1.
    return covtheta


def transform_centroid(xbar, ybar, transform_to='wc'):
    """
    Transform x-bar and y-bar for classical regression fit from Terra-Wasserburg
    to Wetheril or vice versa.
    WC variables are denoted X, Y
    T-W variables are denoted x, y
    """
    assert transform_to in ('wc', 'tw')
    if transform_to == 'wc':
        Xbar = ybar * cfg.U / xbar
        Ybar = 1. / xbar
        return Xbar, Ybar
    elif transform_to == 'tw':
        Xbar, Ybar = xbar, ybar
        xbar = 1 / Ybar
        return Ybar * Xbar / cfg.U


def transform_dp(x0, ox0, y0, oy0, r_xy0, to='wc'):
    """
    Transform data points. E.g. see Ludwig (1998, 2012) and Noda (2017).
    WC --> X, Y
    T-W --> x, y

    """
    assert to in ('wc', 'tw')
    #TODO: use matrix multiplication instead ??
    if to == 'wc':
        x, y = x0, y0
        ox, oy, r_xy = ox0, oy0, r_xy0
        # convert to relative errors
        sx = ox / x
        sy = oy / y
        # do transformation
        Y = 1 / x
        X = cfg.U * y / x
        sX = np.sqrt(sx ** 2 + sy ** 2 - 2 * sx * sy * r_xy)
        sY = sx
        r_XY = (sx ** 2 - sx * sy * r_xy) / (sx * sy)
        # back to abs. errors
        oX = abs(sX * X)
        oY = abs(sY * Y)
        return np.array((X, oX, Y, oY, r_XY))

    else:
        raise ValueError('not yet coded')




"""
"Classical" isochron ages.

"""

import numpy as np

from . import monte_carlo as mc
from . import cfg


diagram_names = {'iso-Pb6U8': '238U-206Pb isochron age',
                 'iso-Pb7U5': '235U-207Pb isochron age'}


def age(b, sb=None, age_type='iso-Pb6U8', dc_errors=False):
    """
    Calculate classical isochron age.
    """
    assert age_type in ('iso-Pb6U8', 'iso-Pb7U5')
    if age_type == 'iso-Pb6U8':
        dc = cfg.lam238
        sdc = cfg.s238
    elif age_type == 'iso-Pb7U5':
        dc = cfg.lam235
        sdc = cfg.s235
    t = 1. / dc * np.log(b + 1.)
    if sb is None:
        return t
    sdc = 0 if dc_errors else sdc
    db = 1. / (dc * (b + 1.))
    ddc = 1. / ((dc ** 2) * np.log(b + 1.)) if dc_errors else 0.
    st = np.sqrt((ddc * sdc) ** 2 + (db * sb) ** 2)
    return t, st


def mc_uncert(fit, age_type='iso-68', dc_errors=False, norm_isotope='204Pb',
              trials=50_000, hist=False):
    """
    Compute classical isochron age uncertainties using Monte Carlo approach.
    """
    flags = np.zeros(trials)
    t = age(fit['theta'][1], age_type='iso-Pb6U8')

    if age_type == 'iso-Pb6U8':
        dc = cfg.lam238
        sdc = cfg.s238
    elif age_type == 'iso-Pb7U5':
        dc = cfg.lam235
        sdc = cfg.s235

    # draw randomised slope / intercept values
    theta, flags = mc.draw_theta(fit, trials, flags)

    # draw randomised decay constant values
    if dc_errors:
        dc = cfg.rng.normal(dc, sdc, trials)

    # simulate ages
    ts = 1. / dc * np.log(theta[1] + 1.)
    flags = mc.update_flags(flags, ts < 0, mc.NEGATIVE_AGE)

    ok = flags == 0
    age_95ci = np.quantile(ts[ok], (0.025, 0.975))
    results = {
        'age_type': diagram_names[age_type],
        'normalising_isotope': norm_isotope,
        'age_1s': np.std(ts[ok]),
        'age_95ci': age_95ci,
        'age_95pm': np.nanmean([t - age_95ci[0], age_95ci[1] - t]),
        'mean_age': np.nanmean(ts[ok]),
        'median_age': np.nanmedian(ts[ok]),
        'y-int_1s': np.nanmean(theta[0][ok]),
        'y-int_95ci': np.nanquantile(theta[0][ok], (0.025, 0.975)),
        'mean_y-int': np.nanmean(theta[0][ok]),
        'median_y-int': np.nanmedian(theta[0][ok]),
        'trials': trials,
        'fails': np.sum(flags != 0),
        'not_converged': np.sum(flags == mc.NON_CONVERGENCE),
        'negative_ages': np.sum(flags == mc.NEGATIVE_AGE)
    }

    if hist:
        fig = mc.age_hist(ts, age_type, a=theta[1], b=theta[0])
        results['age_hist'] = fig

    return results



"""
Monte Carlo disequilibrium U-Pb age errors

"""
import warnings

import numpy as np
import matplotlib.pyplot as plt

from scipy import optimize

from . import minimise
from . import ludwig
from .. import monte_carlo as mc
from .. import cfg
from .. import concordia
from .. import plotting
from .. import useries


#===========================================================================
# Concordia-intercept ages
#===========================================================================

def concint_age(t, fit, A, sA, init, trials=50_000, dc_errors=False,
        diagram='tw', u_errors=False, negative_ar=True, negative_ages=True,
        hist=(False, False), intercept_plot=True,
        intercept_plot_kw=None, conc_kw=None):
    """
    Propagate disequilibrium U-Pb concordia-intercept age uncertainties using
    Monte Carlo simulation.
    
    Parameters
    ----------
    t0 : float
        age solution
    fit : dict
        linear regression fit parameters
    A : np.ndarray
        one-dimensional array of activity ratio values arranged as follows
        - [234U/238U], [230Th/238U], [226Ra/238U], [231Pa/235U]
    sA : np.ndarray
        one-dimensional array of activity ratio value uncertainties given
        as 1 sigma absolute and arranged in the same order as A
    init : array-like
        two-element list of boolean values, the first is True if [234U/238U]
        is an initial value and False if a present-day value, the second is True
        if [230Th/238U] is an initial value and False if a present-day value

    """
    assert diagram == 'tw'  # wc not yet implemented
    flags = np.zeros(trials, dtype='uint8')

    if intercept_plot_kw is None:
        intercept_plot_kw = {}
    if conc_kw is None:
        conc_kw = {}

    theta, flags = mc.draw_theta(fit, trials, flags)
    U = cfg.rng.normal(cfg.U, cfg.sU, trials) if u_errors else cfg.U
    DC8, flags = draw_decay_const(trials, flags, dc_errors=dc_errors)
    DC5, flags = draw_decay_const(trials, flags, dc_errors=dc_errors, 
                                  series='235U')
    BC8 = ludwig.bateman(DC8, series='238U')
    BC5 = ludwig.bateman(DC5, series='235U')

    pA = np.empty((4, trials))
    pA[:3, :], flags = draw_ar(A[:-1], sA[:-1], trials, flags, positive_only=not negative_ar)
    pA[3, :], flags = draw_ar(A[-1], sA[-1], trials, flags, series='235U',
                              positive_only=not negative_ar)

    args = (*theta, pA[:3], pA[-1], DC8, DC5, BC8, BC5, U)

    # run vectorised Newton routine:
    fmin, dfmin = minimise.concint(t0=1.0, diagram='tw', init=init)
    ts, c, zd = optimize.newton(fmin, np.ones(trials) * t, dfmin, args=args,
                                full_output=True)
    flags = mc.check_ages(ts, c, flags, negative_ages=negative_ages)

    # calculate initial activity ratios
    A48i, A08i = useries.Ai_solutions(ts, pA[:3, :], init, DC8)

    if not negative_ar:
        if A48i is not None:
            flags = mc.update_flags(flags, A48i < 0., mc.NEGATIVE_AR_SOL)
        if A08i is not None:
            flags = mc.update_flags(flags, A08i < 0., mc.NEGATIVE_AR_SOL)

    ok = flags == 0
    if sum(ok) == 0:
        raise RuntimeError('all Monte Carlo simulation trials failed')

    age_95ci = np.quantile(ts[ok], (0.025, 0.975))

    # compute results
    results = {
        'age_type': 'concordia-intercept',
        'age_1s': np.std(ts[ok]),
        'age_95ci': age_95ci,
        'age_95pm': np.nanmean([t - age_95ci[0], age_95ci[1] - t]),
        'mean_age': np.nanmean(ts[ok]),
        'median_age': np.nanmedian(ts[ok]),
        'trials': trials,
        'fails': np.sum(flags != 0),
        'not_converged': np.sum(flags == mc.NON_CONVERGENCE),
        'negative_ages': np.sum(flags == mc.NEGATIVE_AGE),
        'negative_ar': np.sum(flags == mc.NEGATIVE_AR_SIM),
        'negative_ar_soln': np.sum(flags == mc.NEGATIVE_AR_SOL)
    }

    if A48i is not None:
        results['mean_[234U/238U]_i'] = np.nanmean(A48i[ok])
        results['median_[234U/238U]_i'] = np.nanmedian(A48i[ok])
        results['[234U/238U]i_1sd'] = np.std(A48i[ok])
        results['[234U/238U]i_95ci'] = np.quantile(A48i[ok], (0.025, 0.975))
        results['cor_age_[234U/238U]i'] = np.corrcoef(np.row_stack((A48i[ok],
                                                A48i[ok])))[0, 1]

    if A08i is not None:
        results['mean_[230Th/238U]_i'] = np.nanmean(A08i[ok])
        results['median_[230Th/238U]_i'] = np.nanmedian(A08i[ok])
        results['[230Th/238U]i_1sd'] = np.nanstd(A08i[ok])
        results['[230Th/238U]i_95ci'] = np.quantile(A08i[ok], (0.025, 0.975))
        results['cor_age_[230Th/238U]i'] = np.corrcoef(np.row_stack((A08i[ok],
                                                 A08i[ok])))[0, 1]

    # compile plots:
    if intercept_plot:
        fig = diseq_intercept_plot(ts, fit, pA[:3, :], pA[3], DC8, DC5, BC8, BC5,
                   U, flags, dp=None, init=init, dc_errors=dc_errors,
                   u_errors=u_errors, **intercept_plot_kw)
        ax = fig.get_axes()[0]
        # TODO: allow concordia func to accept pre-simulated A values to account
        # rejected values...
        concordia.plot_diseq_concordia(ax, A, init, diagram, sA=sA,
                                negative_ar=negative_ar, **conc_kw)
        results['fig'] = fig

    if any(hist):
        if hist[0]:
            a, b = theta
            xs, ys = concordia.diseq_xy(ts, pA, init, diagram)
            fig = mc.age_hist(ts[ok], diagram, a[ok], b[ok], xs[ok], ys[ok])
            results['age_hist'] = fig
        if hist[1]:
            fig = mc.ar_hist(np.transpose(np.transpose(pA[:3])[ok]),
                             pA[-1][ok], init)
            results['ar_hist'] = fig
        if hist[1] and (A48i is not None or A08i is not None):
            if A48i is not None:
                A48i = A48i[ok]
            if A08i is not None:
                A08i = A08i[ok]
            fig = mc.ar_sol_hist(ts[ok], A48i, A08i)
            results['arvi_hist'] = fig

    return results


#===========================================================================
# U-Pb isochron ages
#===========================================================================

def isochron_age(t, fit, A, sA, init=(True, True), trials=50_000,
            dc_errors=False, negative_ar=True, negative_ages=True,
            hist=(False, False), age_type='iso-Pb6U8'):
    """
    Propogate disequilibrium isochron age uncertainties using Monte Carlo
    approach.
    """
    assert age_type in ('iso-Pb6U8', 'iso-Pb7U5')
    flags = np.zeros(trials, dtype='uint8')

    theta, flags = mc.draw_theta(fit, trials, flags)

    if age_type == 'iso-Pb6U8':
        pA, flags = draw_ar(A, sA, trials, flags, positive_only=not negative_ar)
        DC, flags = draw_decay_const(trials, flags, dc_errors=dc_errors)
        BC = ludwig.bateman(DC, series='238U')
    else:
        pA, flags = draw_ar(A, sA, trials, flags, series='235U',
                            positive_only=not negative_ar)
        DC, flags = draw_decay_const(trials, flags, dc_errors=dc_errors, series='235U')
        BC = ludwig.bateman(DC, series='235U')
    # compile args for numercial age routine:
    args = (theta[1], pA, DC, BC)

    # run vectorised Newton routine:
    fmin, dfmin = minimise.isochron(t0=t, age_type=age_type, init=init)
    ts, c, zd = optimize.newton(fmin, np.ones(trials) * t, dfmin, args=args,
                                full_output=True)
    flags = mc.check_ages(ts, c, flags, negative_ages=negative_ages)

    # back-calculate activity ratios
    if age_type == 'iso-Pb6U8':
        A48i, A08i = useries.Ai_solutions(ts, pA, init, DC)
        if not negative_ar:
            if A48i is not None:
                flags = mc.update_flags(flags, A48i < 0., mc.NEGATIVE_AR_SOL)
            if A08i is not None:
                flags = mc.update_flags(flags, A08i < 0., mc.NEGATIVE_AR_SOL)

    ok = flags == 0

    if sum(ok) == 0:
        raise RuntimeError('all Monte Carlo simulation trials failed')

    age_95ci = np.quantile(ts[ok], (0.025, 0.975))

    # compute results
    results = {
        'age_type': f'{age_type} isochron age',
        'age_1s': np.nanstd(ts[ok]),
        'age_95ci': age_95ci,
        'age_95pm': np.nanmean((t - age_95ci[0], age_95ci[1] - t)),
        'mean_age': np.nanmean(ts[ok]),
        'median_age': np.nanmedian(ts[ok]),
        'trials': trials,
        'fails': np.sum(flags != 0),
        'not_converged': np.sum(flags == mc.NON_CONVERGENCE),
        'negative_ages': np.sum(flags == mc.NEGATIVE_AGE),
        'negative_ar': np.sum(flags == mc.NEGATIVE_AR_SIM),
        'negative_ar_soln': np.sum(flags == mc.NEGATIVE_AR_SOL)
    }

    if age_type == 'iso-Pb6U8':
        if A48i is not None:
            results['mean_[234U/238U]i'] = np.nanmean(A48i[ok])
            results['median_[234U/238U]i'] = np.nanmedian(A48i[ok])
            results['[234U/238U]i_1sd'] = np.nanstd(A48i[ok])
            results['[234U/238U]i_95ci'] = np.quantile(A48i[ok], (0.025, 0.975))
            results['cor_[234U/238U]i_t'] = \
                np.corrcoef(np.row_stack((ts[ok], A48i[ok])))[0, 1]

        if A08i is not None:
            results['mean_init_[230Th/238U]'] = np.nanmean(A08i[ok])
            results['median_init_[230Th/238U]'] = np.nanmedian(A08i[ok])
            results['[230Th/238U]i_1sd'] = np.nanstd(A08i[ok])
            results['[230Th/238U]i_95ci'] = np.quantile(A08i[ok], (0.025, 0.975))
            results['cor_[230Th/238U]i_t'] = \
                np.corrcoef(np.row_stack((ts[ok], A08i[ok])))[0, 1]

    # compile plots
    if any(hist):
        if hist[0]:
            a, b = theta
            fig = mc.age_hist(ts[ok], age_type, a[ok], b[ok])
            results['age_hist'] = fig
        if hist[1]:
            A238 = None if age_type != 'iso-Pb6U8' else pA.T[ok].T
            A235 = None if age_type != 'iso-Pb7U5' else pA.T[ok].T
            fig = mc.ar_hist(A238, A235, init=init)
            results['ar_hist'] = fig
        if age_type == 'iso-Pb6U8':
            if hist[1] and (not init[0] or not init[1]):
                if not init[0]:
                    A48i = A48i[ok]
                if not init[1]:
                    A08i = A08i[ok]
                fig = mc.ar_sol_hist(ts[ok], A48i, A08i)
                results['arvi_hist'] = fig

    return results


#===========================================================================
# Forced-concordance [234U/238U] errors
#===========================================================================

def fc_A48i(t57, A48i, fit_57, fit_86, A, sA, init, trials=50_000,
            negative_ar=True, negative_ages=True, hist=(0,0,0)):
    """
    Forced-concordance initial [234U/238U] value errors.
    """
    assert init[1]
    flags = np.zeros(trials)
    theta57 = cfg.rng.multivariate_normal(fit_57['theta'], fit_57['covtheta'], trials)
    theta86 = cfg.rng.multivariate_normal(fit_86['theta'], fit_86['covtheta'], trials)
    
    pA = np.empty((4, trials))
    pA[-1, :], flags = draw_ar(A[-1], sA[-1], trials, flags, series='235U',
                               positive_only=not negative_ar)
    pA[:3, :], flags = draw_ar(A[:-1], sA[:-1], trials, flags, positive_only=not negative_ar)

    DC8, flags = draw_decay_const(trials, flags, dc_errors=False)
    DC5, flags = draw_decay_const(trials, flags, dc_errors=False, series='235U')
    BC8 = ludwig.bateman(DC8, series='238U')
    BC5 = ludwig.bateman(DC8, series='235U')

    args57 = (theta57[:, 1], pA[-1], DC5, BC5)
    
    # run vectorised Newton routine to get iso-57 ages
    fmin, dfmin = minimise.isochron(t0=1.0, age_type='iso-Pb7U5')
    ts, c, zd = optimize.newton(fmin, np.ones(trials) * t57, dfmin, args=args57,
                                full_output=True)
    flags = mc.check_ages(ts, c, flags, negative_ages=negative_ages)

    # run vectorised Newton routine to get initial [234U/238U] values:
    fmin, dfmin = minimise.concordant_A48()
    A48i_s, c, zd = optimize.newton(fmin, np.full(trials, A48i), dfmin,
            args=(ts, theta86[:, 1], pA[:3], DC8, BC8), full_output=True)
    flags = mc.check_ages(A48i_s, c, flags, negative_ages=negative_ar) # pretend ar48i solutions are ages

    ok = (flags == 0)
    if sum(ok) == 0:
        raise RuntimeError('all Monte Carlo simulation trials failed')
    age_95ci = np.quantile(ts[ok], (0.025, 0.975))
    A48i_95ci = np.quantile(A48i_s[ok], (0.025, 0.975))

    # compute results
    results = {
        'age_type': 'forced-concordance [234U/238U]_i',
        '207Pb_age_1sd': np.nanstd(ts[ok]),
        '207Pb_age_95ci': age_95ci,
        '207Pb_age_95pm': [np.nanmean((age_95ci[1] - t57, t57 - age_95ci[0]))],
        'mean_207Pb_age': np.nanmean(ts[ok]),
        'median_207Pb_age': np.nanmedian(ts[ok]),
        '[234U/238U]i_1sd': np.nanstd([A48i_s[ok]]),
        '[234U/238U]i_95ci': A48i_95ci,
        '[234U/238U]i_95pm': [np.nanmean((A48i_95ci[1] - A48i, A48i - A48i_95ci[0]))],
        'mean_[234U/238U]i': np.nanmean([A48i_s[ok]]),
        'median_[234U/238U]i': np.nanmedian([A48i_s[ok]]),
        'trials': trials,
        'fails': np.sum(flags != 0),
        'not_converged': np.sum(flags == mc.NON_CONVERGENCE),
        'negative_ages': np.sum(flags == mc.NEGATIVE_AGE),
        'negative_ar': np.sum(flags == mc.NEGATIVE_AR_SIM)
    }

    if any(hist):
        if hist[0]:
            a, b = np.transpose(theta57)
            fig = mc.age_hist(ts[ok], 'iso-Pb7U5', a[ok], b[ok])
            results['age_hist'] = fig
        if hist[1]:
            A238 = [A48i_s[ok], *pA[1:3].T[ok].T]
            A235 = pA[-1].T[ok].T
            fig = mc.ar_hist(A238, A235)
            results['ar_hist'] = fig

    return results


#===========================================================================
# Pb/U ages
#===========================================================================

def pbu_age(t, x, sx, trials=10_000, DThU=None, DThU_1s=None, DPaU=None,
            DPaU_1s=None, ThU_min=None, ThU_melt=None, ThU_melt_1s=None,
            DThU_const=True, method='Ludwig', age_type='Pb6U8',
            negative_ages=True, negative_ar=True, rand=False,
            hist=(False, False)):
    """
    Propagate disequilibrium Pb/U age uncertainties using Monte Carlo approach.

    If rand is True, will also simulate
    random only age uncertainties. Expects initial activity ratio values and
    does not yet account for activity ratio errors.

    Parameters
    ----------
    x :
        Pb/U ratios
    sx :
        Pb/U ratio errors (1s, abs.)

    Notes
    -----
    Does not account for decay constant errors.
    negative_ar option not yet implemented

    """
    assert age_type in ('Pb6U8', 'Pb7U5')
    assert method in ('Ludwig', 'Guillong')
    assert not any([np.isnan(t) for t in t])       # nan ages should be removed
    if negative_ar:
        warnings.warn('negative_ar option not yet implemented in Monte Carlo simulation')

    A48_eq = 1.0
    A68_eq = 1.0

    n = x.size
    flags = np.zeros((trials, n), dtype='uint8')
    ts = np.empty((trials, n))                  # pre-allocate ages array
    if rand:
        flagr = np.zeros((trials, n), dtype='uint8')
        tr = np.empty((trials, n))             # random only uncertainties

    # simulate fXU values
    if age_type == 'Pb6U8':
        if DThU_const:
            fThU = np.random.normal(DThU, DThU_1s, trials)
            fXU = np.broadcast_to(fThU.reshape(trials, 1), (trials, n))
        else:
            ThU_melt_s = np.random.normal(ThU_melt, ThU_melt_1s, (trials, 1))
            fXU = np.broadcast_to(ThU_min, (trials, n)) / ThU_melt_s

    elif age_type == 'Pb7U5':
        fPaU = np.random.normal(DPaU, DPaU_1s, trials)
        fXU = np.broadcast_to(fPaU.reshape(trials, 1), (trials, n))

    if method == 'Ludwig':
        fmin, dfmin = minimise.pbu(age_type=age_type, t0=np.nanmean(t))
        if age_type == 'Pb6U8':
            DC = (cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226)
            BC = ludwig.bateman(DC)
        else:
            DC = (cfg.lam235, cfg.lam231, cfg.lam227)
            BC = ludwig.bateman(DC, series='235U')

    elif method == 'Guillong':
        fmin, dfmin = minimise.guillong(t0=np.nanmean(t))

    # Do rand + systematic age simulation for each age data point:
    for i in range(n):
        xs = cfg.rng.normal(x[i], sx[i], trials)

        # simulate activity ratios
        if method == 'Ludwig':
            if age_type == 'Pb6U8':
                args = (xs, (A48_eq, fXU[:, i], A68_eq), DC, BC)
            else:
                args = (xs, (fXU[:, i]), DC, BC)
        elif method == 'Guillong':
            if age_type == 'Pb6U8':
                args = (xs, fXU[:, i], cfg.lam238, cfg.lam230)
            else:
                args = (xs, fXU[:, i], cfg.lam235, cfg.lam231)

        t0 = np.full(trials, t[i])
        r, c, zd = optimize.newton(fmin, t0, dfmin, args=args, full_output=True)
        flags[:, i] = mc.check_ages(r, np.logical_or(c, ~zd), flags[:, i],
                negative_ages=negative_ages)
        ts[:, i] = r

        # Compute random only errors (for plotting purposes)
        if rand:
            if age_type == 'Pb6U8':
                if DThU_const:
                    rfXU = DThU
                else:
                    rfXU = ThU_min[i] / ThU_melt
            elif age_type == 'Pb7U5':
                rfXU = np.atleast_1d(DPaU)

            if method == 'Ludwig':
                if age_type == 'Pb6U8':
                    DC = (cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226)
                    BC = ludwig.bateman(DC)
                    args = (xs, [A48_eq, rfXU, A68_eq], DC, BC)
                else:
                    DC = (cfg.lam235, cfg.lam231)
                    BC = ludwig.bateman(DC, series='235U')
                    args = (xs, rfXU, DC, BC)

            elif method == 'Guillong':
                if age_type == 'Pb6U8':
                    args = (xs, rfXU, cfg.lam238, cfg.lam230)
                else:
                    args = (xs, rfXU, cfg.lam235, cfg.lam231)

            r, c, zd = optimize.newton(fmin, t0, dfmin, args=args, full_output=True)
            flagr[:, i] = mc.check_ages(r, np.logical_or(c, ~zd), flags[:, i],
                                     negative_ages=negative_ages)
            tr[:, i] = r

    ok = np.all(flags == 0, axis=1)
    age_95ci = [np.quantile(t[ok], (0.025, 0.975)) for t in ts.T]
    # compile results
    results = {
        'age_type': age_type,
        'age_1s': [np.nanstd(t[ok]) for t in ts.T],
        'age_95pm': [np.nanmean((t - m, p - t)) for m, p in age_95ci],
        'age_95ci': age_95ci,
        'mean_age': [np.nanmean(t[ok]) for t in ts.T],
        'median_age': [np.nanmedian(t[ok]) for t in ts.T],
        'cov_t': np.cov(np.transpose(ts[ok])),
        'trials': trials,
        'fails': np.sum(flags != 0, axis=1),
        'not_converged': np.sum(flags == mc.NON_CONVERGENCE, axis=1),
        'negative_ages': np.sum(flags == mc.NEGATIVE_AGE, axis=1),
        'negative_distr_coeff': np.sum(flags == mc.NEGATIVE_AR_SIM, axis=1)
    }
    if rand:
        okr = np.all(flagr == 0, axis=1)
        rand_age_95ci = [np.quantile(t[okr], (0.025, 0.975)) for t in tr.T]
        results['age_rand_1s'] = [np.nanstd(t[okr]) for t in tr.T]
        results['age_rand_95ci'] = rand_age_95ci
        results['age_rand_95pm'] = [np.nanmean((t - m, p - t)) for m, p in rand_age_95ci]
        results['mean_rand_age'] = [np.nanmean(t[okr]) for t in tr.T]
        results['median_rand_age'] = [np.nanmedian(t[okr]) for t in tr.T]
        results['rand_fails'] = np.sum(flagr != 0, axis=1)
        results['rand_not_converged'] = np.sum(flagr == mc.NON_CONVERGENCE, axis=1)
        results['rand_negative_ages'] = np.sum(flagr == mc.NEGATIVE_AGE, axis=1)
        results['rand_distr_coeff'] = np.sum(flagr == mc.NEGATIVE_AR_SIM, axis=1)

    if any(hist):
        raise ValueError('not yet coded')

    return results


def mod207_age(t, x, sx, y, sy, r_xy, Pb76, Pb76se, ThU_min=None,
               ThU_melt=None, ThU_melt_1s=None, DThU=None, DThU_1s = None,
               DPaU = None, DPaU_1s = None, DThU_const = True, trials=10_000,
               method='Ludwig', negative_ages=True, negative_ar=True,
               rand=False, hist=(False, False)):
    """
    Compute Pb/U age uncertainties using Monte Carlo approach. Does not
    account for decay constant errors. If rand is True, will also simulate
    random only age uncertainties. Expects initial activity ratio values and
    does not yet account for activity ratio errors.

    U and decay constant errors not included.
    """
    assert method in ('Ludwig', 'Sakata')
    assert not any([np.isnan(t) for t in t])       # nan ages should be removed
    if negative_ar:
        warnings.warn('negative_ar option not yet implemented in Monte Carlo simulation')

    A48_eq = 1.0
    A68_eq = 1.0
    init = (True, True)

    n = x.size
    flags = np.zeros((trials, n), dtype='uint8')
    ts = np.empty((trials, n))                  # pre-allocate ages array
    if rand:
        flagr = np.zeros((trials, n), dtype='uint8')
        tr = np.empty((trials, n))             # random only uncertainties

    # simulate fXU values
    if DThU_const:
        fThU = np.random.normal(DThU, DThU_1s, trials)
        fThU = np.broadcast_to(fThU.reshape(trials, 1), (trials, n))
    else:
        ThU_melt_s = np.random.normal(ThU_melt, ThU_melt_1s, (trials, 1))
        fThU = np.broadcast_to(ThU_min, (trials, n)) / ThU_melt_s

    fPaU = np.random.normal(DPaU, DPaU_1s, trials)
    fPaU = np.broadcast_to(fPaU.reshape(trials, 1), (trials, n))

    flags = mc.update_flags(flags, fThU < 0., mc.NEGATIVE_AR_SIM)
    flags = mc.update_flags(flags, fPaU < 0., mc.NEGATIVE_AR_SIM)

    # simulate U and Pb76 values
    Pb76 = cfg.rng.normal(Pb76, Pb76se, trials)
    U = cfg.U

    if method == 'Ludwig':
        fmin, dfmin = minimise.mod207(t0=np.nanmean(t))
        DC8 = (cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226)
        DC5 = (cfg.lam235, cfg.lam231)
        BC8 = ludwig.bateman(DC8)
        BC5 = ludwig.bateman(DC5, series='235U')

    elif method == 'Sakata':
        fmin, dfmin = minimise.sakata(t0=np.nanmean(t))

    # Initialise arrays for storing histograms:
    if hist[0]: age_hist = []
    if hist[1]: arv_hist = []

    # Do random + systematic age simulation for each age data point:
    for i in range(n):
        mu = np.array([x[i], y[i]])
        cov = np.array([[sx[i] ** 2, r_xy[i] * sx[i] * sy[i]],
                        [r_xy[i] * sx[i] * sy[i], sy[i] ** 2]])
        dp = cfg.rng.multivariate_normal(mu, cov, trials)
        xs, ys = dp[:, 0], dp[:, 1]

        # simulate activity ratios
        if method == 'Ludwig':
            args = (xs, ys, [A48_eq, fThU[:, i], A68_eq, fPaU[:, i]],
                    DC8, DC5, BC8, BC5, U, Pb76)
        elif method == 'Sakata':
            args = (xs, ys, fThU[:, i], fPaU[:, i], cfg.lam238, cfg.lam230,
                    cfg.lam235, cfg.lam231, U, Pb76)

        t0 = np.full(trials, t[i])
        r, c, zd = optimize.newton(fmin, t0, dfmin, args=args, full_output=True)
        flags[:, i] = mc.check_ages(r, np.logical_or(c, ~zd), flags[:, i],
                                    negative_ages=negative_ages)
        ts[:, i] = r

        # TODO: check for reverse discordance?

        # make histograms
        if hist[0]:
            fig = mc.age_hist(ts[:, i], flags[:, i], 'mod207Pb', x=xs, y=ys)
            age_hist.append(fig)
        if hist[1]:
            warnings.warn('ar hist option not yet implemented')

        # Random errors only (e.g. for plotting)
        if rand:
            if DThU_const:
                rfThU = DThU
                rfThU = np.broadcast_to(rfThU, (n,))
            else:
                rfThU = ThU_min / ThU_melt

            rfPaU = np.atleast_1d(DPaU)

            # no need to check for negative distr. coefficients here since
            # they have no error attached

            if method == 'Ludwig':
                args = (xs, ys, [A48_eq, rfThU[i], A68_eq, rfPaU],
                        DC8, DC5, BC8, BC5, U, Pb76)
            elif method == 'Sakata':
                args = (xs, ys, rfThU[i], rfPaU, cfg.lam238, cfg.lam230,
                        cfg.lam235, cfg.lam231, U, Pb76)

            r, c, zd = optimize.newton(fmin, t0, dfmin, args=args, full_output=True)
            flagr[:, i] = mc.check_ages(r, np.logical_or(c, ~zd), flags[:, i],
                                        negative_ages=negative_ages)
            tr[:, i] = r



    # compile results
    ok = np.all(flags == 0, axis=1)
    results = {
        'age_type': 'Modified 207Pb',
        'method': method,
        'age_1s': [np.nanstd(t[ok]) for t in ts.T],
        'age_95ci': [np.quantile(t[ok], (0.025, 0.975)) for t in ts.T],
        'mean_age': [np.nanmean(t[ok]) for t in ts.T],
        'median_age': [np.nanmedian(t[ok]) for t in ts.T],
        'cov_t': np.cov(np.transpose(ts[ok])),
        'trials': trials,
        'fails': np.sum(flags != 0, axis=1),
        'not_converged': np.sum(flags == mc.NON_CONVERGENCE, axis=1),
        'negative_ages': np.sum(flags == mc.NEGATIVE_AGE, axis=1),
        'negative_distr_coeff': np.sum(flags == mc.NEGATIVE_AR_SIM, axis=1),
    }
    if rand:
        okr = np.all(flagr == 0, axis=1)
        results['age_rand_1s'] = [np.nanstd(t[okr]) for t in tr.T]
        results['age_rand_95ci'] = [np.quantile(t[okr], (0.025, 0.975)) for t in tr.T]
        results['mean_rand_age'] = [np.nanmean(t[okr]) for t in tr.T]
        results['median_rand_age'] = [np.nanmedian(t[okr]) for t in tr.T]
        results['rand_fails'] = np.sum(flagr != 0, axis=1)
        results['rand_not_converged'] = np.sum(flagr == mc.NON_CONVERGENCE, axis=1)
        results['rand_negative_ages'] = np.sum(flagr == mc.NEGATIVE_AGE, axis=1)
        results['rand_distr_coeff'] = np.sum(flagr == mc.NEGATIVE_AR_SIM, axis=1)

    if hist[0]:
        results['age_hist'] = age_hist
    if hist[1]:
        results['ar_hist'] = None

    return results


#===============================================
# Simulate constants and activity ratios
#===============================================

def draw_ar(A, sA, trials, flags=None, positive_only=False,
            series='238U'):
    """  
    Draw random activity ratio values.
    
    """
    assert series in ('238U', '235U')
    
    if series == '238U':
        A48 = cfg.rng.normal(A[0], sA[0], trials)
        A08 = cfg.rng.normal(A[1], sA[1], trials)
        A68 = cfg.rng.normal(A[2], sA[2], trials)
        if flags is not None and positive_only:
            for x in (A48, A08, A68):
                flags = mc.update_flags(flags, x < 0., mc.NEGATIVE_AR_SIM)
        pA = np.array((A48, A08, A68))
    
    else:
        pA = cfg.rng.normal(A, sA, trials)
        if positive_only:
            flags = mc.update_flags(flags, pA < 0., mc.NEGATIVE_AR_SIM)
    
    return pA, flags


def draw_decay_const(trials, flags, dc_errors=False, series='238U', cor=False):
    """  
    Draw random decay constant values.

    """
    assert series in ('238U', '235U')
    
    if series == '238U':
        # note: effects of error correlations are probably always negligible ?
        if dc_errors:
            if cor:
                cov_84 = cfg.cor_238_234 * (cfg.s238 * cfg.s234)
                cov_80 = cfg.cor_238_230 * (cfg.s238 * cfg.s230)
                cov_40 = cfg.cor_234_230 * (cfg.s234 * cfg.s230)
    
                V = np.array([[cfg.s238 ** 2, cov_84, cov_80, 0.],
                              [cov_84, cfg.s234 ** 2, cov_40, 0.],
                              [cov_80, cov_40, cfg.s230 ** 2, 0.],
                              [0., 0., 0., cfg.s226 ** 2]])
    
                DC = cfg.rng.multivariate_normal(
                    (cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226), V, trials
                )

            else:
                DC = cfg.rng.normal((cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226),
                                    (cfg.s238, cfg.s234, cfg.s230, cfg.s226),
                                    (trials, 4))
            DC = np.transpose(DC)
        else:
            DC = np.array((cfg.lam238, cfg.lam234, cfg.lam230, cfg.lam226))
    
    else:
        if dc_errors:
            lam235 = cfg.rng.normal(cfg.lam235, cfg.s235, trials)
            lam231 = cfg.rng.normal(cfg.lam231, cfg.s231, trials)
            DC = np.array((lam235, lam231))
        else:
            DC = np.array((cfg.lam235, cfg.lam231))
    
    return DC, flags


#===============================================
# Plots
#===============================================

def diseq_intercept_plot(ts, fit, A8, A15, DC8, DC5, BC8, BC5, U, flags,
            init=(True, True), dp=None, dc_errors=False, u_errors=False,
            diagram='tw', xlim=(None, None), ylim=(None, None),
            intercept_points=True, intercept_ellipse=False):
    """
    Plot simulated disequilibrium concordia intercept points.
    """
    fig, ax = plt.subplots(**cfg.fig_kw, subplot_kw=cfg.subplot_kw)
    ax = fig.axes[0]
    ok = flags == 0

    # Filter out failed iteration values:
    if dc_errors:
        DC8 = np.transpose(np.transpose(DC8)[ok])
        DC5 = np.transpose(np.transpose(DC5)[ok])
        BC8 = np.transpose(np.transpose(BC8)[ok])
        BC5 = np.transpose(np.transpose(BC5)[ok])
    if u_errors:
        U = U[ok]

    A8 = np.transpose(np.transpose(A8)[ok])
    A15 = A15[ok]

    x = 1. / ludwig.f(ts[ok], A8, DC8, BC8, init=init)
    y = ludwig.g(ts[ok], A15, DC5, BC5) * x / U
    mc.intercept_plot_axis_limits(ax, x, y, diagram=diagram)

    if intercept_points:
        ax.plot(x, y, **cfg.conc_intercept_markers_kw)

    if intercept_ellipse:
        cov = np.cov(x, y)
        e = plotting.confidence_ellipse2(ax, np.nanmean(x), np.nanmean(y), cov,
                     **cfg.conc_intercept_ellipse_kw)
        ax.add_patch(e)

    # Plot data ellipses if dp given:
    if dp is not None:
        plotting.plot_dp(ax, *dp, reset_axis_limits=False)

    # Label axes and apply plot settings etc.
    plotting.apply_plot_settings(fig, plot_type='intercept', diagram=diagram,
                xlim=xlim, ylim=ylim, tight_layout=True)
    # Plot regression line and envelope
    plotting.rline(ax, fit['theta'])
    plotting.renv(ax, fit)

    return fig




"""
Monte Carlo error propagation functions

"""

import numpy as np
import matplotlib.pyplot as plt

from . import cfg
from . import plotting


# Fail codes
NON_CONVERGENCE = 1
NEGATIVE_AGE = 2
NEGATIVE_AR_SIM = 3
NEGATIVE_AR_SOL = 4


def draw_theta(fit, trials, flags):
    """
    Draw random linear regression slope and intercept values.
     """
    if fit['type'] == 'classical' and fit['excess_scatter']:
        df = fit['n'] - 2
        theta = multivariate_t_rvs(fit['theta'], fit['covtheta'], df, trials)
    else:
        theta = cfg.rng.multivariate_normal(
            fit['theta'], fit['covtheta'], trials
        )
    return np.transpose(theta), flags


def multivariate_t_rvs(m, S, df, n):
    """
    Generate random variables of multivariate t distribution

    Code based on:
    <https://github.com/statsmodels/statsmodels/blob/master/statsmodels/
    sandbox/distributions/multivariate.py#L90>
    Originally written by Enzo Michelangeli, style changes by josef-pktd.

    Parameters
    ----------
    m : array_like
        mean of random variable, length determines dimension of random variable
    S : array_like
        square array of covariance  matrix
    df : int or float
        degrees of freedom
    n : int
        number of observations, return random array will be (n, len(m))
    Returns
    -------
    rvs : ndarray, (n, len(m))
        each row is an independent draw of a multivariate t distributed
        random variable
    """
    m = np.asarray(m)
    d = len(m)

    if df == np.inf:
        x = 1.
    else:
        x = cfg.rng.chisquare(df, n) / df
    z = cfg.rng.multivariate_normal(np.zeros(d), S, (n,))
    # same output format as random.multivariate_normal
    return m + z / np.sqrt(x)[:, None]


def update_flags(flags, cond, code_num):
    """
    Update array of failed iteration flags.
    """
    if len(cond) != len(flags):
        raise ValueError('flags0, cond, and msk  must be of equal length')
    # Do not overwrite code if iteration is already flagged.
    cond = np.logical_and(cond, flags == 0)
    flags = np.where(cond, code_num, flags)
    return flags


def check_ages(t, c, flags, negative_ages=False):
    """
    Check simulated ages have converged properly.

    Parameters
    ----------
    t : np.ndarray
        Simulated ages.
    c : np.ndarray, logical
        Convergence status.
    """
    # check convergence:
    flags = update_flags(flags, np.logical_or(~c, np.isnan(t)), NON_CONVERGENCE)
    # check negative ages:
    if not negative_ages:
        flags = update_flags(flags, t < 0., NEGATIVE_AGE)
    return flags


def summary_stats(x, p=0.95):
    """Compute summary statistics from array of Monte Carlo simulated values.
    """
    p = p * 100  # convert to percent
    x_av = np.nanmean(x)
    sx = np.nanstd(x)

    # Calcaulte upper and lower limits on coverage interval.
    x_lower = np.nanpercentile(x, (100 - p) / 2)
    x_upper = np.nanpercentile(x, (100 - p) / 2 + p)
    x_pm = np.nanmean((x_upper - x_av, x_av - x_lower))

    return x_av, sx, (x_lower, x_upper), x_pm


def histogram(ax, y, xlabel=None):
    """
    Plot histogram of mc simulated values.
    """
    n = int(np.ceil(np.sqrt(y.shape[0])))
    n = n if n < 1000 else 1000
    ax.hist(y, bins=n, density=True, **cfg.hist_bars_kw)
    ax.set_xlabel(xlabel)

    # if plot_gaussian_fit:
    #     mu = np.nanmean(y)
    #     sigma = np.nanstd(y)
    #     xx = np.linspace(*ax.get_xlim(), 300)
    #     if sigma != 0.:
    #         ax.plot(xx, stats.norm.pdf(xx, mu, sigma),
    #                 **HistOpt.gaussian_line_kw)


def scatterplot(ax, x, y, xlabel=None, ylabel=None):
    """
    Create scatterplot of correlated mc simulated values.
    """
    ax.plot(x, y, **cfg.scatter_markers_kw)
    # reset axis limits as default doesn't always seem to work very well...
    xmin, xmax = x.min(), x.max()
    dx = xmax - xmin
    ymin, ymax = y.min(), y.max()
    dy = ymax - ymin
    ax.set_xlim((xmin - 0.1 * dx, xmax + 0.1 * dx))
    ax.set_ylim((ymin - 0.1 * dy, ymax + 0.1 * dy))
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    plotting.apply_plot_settings(ax.get_figure(), plot_type='hist',
            tight_layout=True, diagram=None)


def age_hist(ts, age_type, a=None, b=None, x=None, y=None):
    """
    Create histogram plots of simulated regression parameters and ages.
    """
    assert age_type in ('tw', 'wc', 'iso-Pb6U8', 'iso-Pb7U5', 'Pb6U8', 'Pb7U5',
                        'mod207Pb')

    fig, ax = plt.subplots(2, 2, **cfg.fig_kw)

    histogram(ax[0, 0], ts, xlabel='age (Ma)')
    if age_type in ('tw', 'wc') or age_type.startswith('iso'):
        histogram(ax[1, 0], a, xlabel='y-int.')
        histogram(ax[0, 1], b, xlabel='slope')
        scatterplot(ax[1, 1], b, a, xlabel='slope', ylabel='y-int.')
    elif age_type == 'Pb6U8':
        histogram(ax[1, 0], x, xlabel='$^{206}$Pb/$^{238}$U')
    elif age_type == 'Pb7U5':
        histogram(ax[1, 0], x, xlabel='$^{207}$Pb/$^{235}$U')
    elif age_type == 'mod207Pb':
        histogram(ax[1, 0], x, xlabel='$^{206}$Pb/$^{238}$U')
        histogram(ax[0, 1], y, xlabel='$^{207}$Pb/$^{235}$U')
        scatterplot(ax[1, 1], x, y, xlabel='$^{206}$Pb/$^{238}$U',
                    ylabel='$^{207}$Pb/$^{235}$U')

    for a in ax.flatten():
        plotting.apply_plot_settings(a.get_figure(), plot_type='hist',
                tight_layout=True, diagram=None)
    fig.tight_layout()

    return fig


def fXU_hist(age_type, flags, fThU=None, fPaU=None,):
    """
    Plot simulated fractionation factors for Guillong and Sakata ages.
    """
    assert (fThU is not None) or (fPaU is not None)
    good = flags == 0
    fig, ax = plt.subplots(2, 2, **cfg.fig_kw)
    ax = ax.flatten()
    i = 0
    if fThU is not None:
        histogram(ax[i], fThU[good], xlabel="fThU")
        i += 1
    if fPaU is not None:
        histogram(ax[i], fPaU[good], xlabel='fPaU')
    return fig


def ar_hist(A238, A235, init=(True, True)):
    """
    Create histogram of simulated activity ratio values for successful
    iterations.
    """
    fig, ax = plt.subplots(2, 2, **cfg.fig_kw)
    axx = ax.flatten()
    i = 0
    if A238 is not None:
        histogram(axx[i], A238[0])
        axx[i].set_xlabel("%s\n[$^{234}$U/$^{238}$U]" %
                          ('initial' if init[0] else 'present-day'))
        i += 1
        histogram(axx[i], A238[1])
        axx[i].set_xlabel("%s\n[$^{230}$Th/$^{238}$U]" %
                          ('initial' if init[1] else 'present-day'))
        i += 1
        histogram(axx[i], A238[2])
        axx[i].set_xlabel("initial\n[$^{226}$Ra/$^{238}$U]")
        i += 1
    if A235 is not None:
        histogram(axx[i], A235)
        axx[i].set_xlabel("initial\n[$^{231}$Pa/$^{235}$U]")

    for a in ax.flatten():
        plotting.apply_plot_settings(fig, plot_type='hist',
                     norm_isotope='204Pb', diagram=None)
    return fig


def ar_sol_hist(t, A48_init, A08_init):
    """
    Create histograms of initial activity ratio solutions.
    """
    fig, ax = plt.subplots(2, 2, **cfg.fig_kw)
    i = 0
    if A48_init is not None:
        histogram(ax[i][0], A48_init)
        ax[i][0].set_xlabel("initial\n[$^{234}$U/$^{238}$U] soln.")
        scatterplot(ax[i][1], A48_init, t)
        ax[i][1].set_ylabel('age (Ma)')
        ax[i][1].set_xlabel("initial\n[$^{234}$U/$^{238}$U] soln.")
        i += 1
    if A08_init is not None:
        histogram(ax[i][0], A08_init)
        ax[i][0].set_xlabel("initial\n[$^{230}$Th/$^{238}$U] soln.")
        scatterplot(ax[i][1], A08_init, t)
        ax[i][1].set_ylabel('age (Ma)')
        ax[i][1].set_xlabel("initial\n[$^{230}$Th/$^{238}$U] soln.")
        i += 1

    for a in ax.flatten():
        plotting.apply_plot_settings(fig, plot_type='isochron',
                     norm_isotope='204Pb', diagram=None)
    return fig


def intercept_plot_axis_limits(ax, x, y, diagram='tw', min_yspan=1e-03,
               tw_f=((3, 3), (2, 5)), wc_f=((3, 3), (3, 3))):
    """
    Naive estimate of reasonable concordia intercept axis limits.
    """
    f = tw_f if diagram == 'tw' else wc_f
    ymin, ymed, ymax = np.quantile(y, (0.05, 0.50, 0.95))
    xmin, xmed, xmax = np.quantile(x, (0.05, 0.50, 0.95))
    yspan = ymax - ymin
    xspan = xmax - xmin

    if diagram == 'tw' and yspan < min_yspan:
        yspan = min_yspan

    xmin = xmin - f[0][0] * xspan
    xmax = xmax + f[0][1] * xspan
    ymin = ymin - f[1][0] * yspan
    ymax = ymax + f[1][1] * yspan

    ax.set_ylim(ymin, ymax)
    ax.set_xlim(xmin, xmax)


def intercept_plot_ax_limits(ax, b, x, y, diagram='tw', min_yspan = 0.001,
                             maxiter=100):
    """
    Compute intercept plot axis limits.
    """
    ymin, ymed, ymax = np.quantile(y, (0.025, 0.50, 0.975))
    xmin, xmed, xmax = np.quantile(x, (0.025, 0.50, 0.975))
    yspan = ymax - ymin
    xspan = xmax - xmin

    dx = 3 * xspan
    ax.set_xlim((xmin - dx, xmax + dx))

    i = 0
    fu = 6 if diagram == 'tw' else 3
    fl = 3
    while (ymax - ymin) < min_yspan and i < maxiter:
        i += 1
        yspan *= 2
        ymin = ymin - fl * yspan
        ymax = ymax + fu * yspan

    ax.set_ylim((ymin, ymax))



"""
Uranium-series functions and routines.

References
----------
.. [Cheng2000]
    Cheng, H., Adkins, J., Edwards, R.L., Boyle, E.A., 2000. U-Th dating of
    deep-sea corals. Geochimica et Cosmochimica Acta 64, 2401–2416.
    https://doi.org/10.1016/S0016-7037(99)00422-6
.. [Hellstrom2006]
    Hellstrom, J., 2006. U–Th dating of speleothems with high initial 230Th using
    stratigraphical constraint. Quaternary Geochronology 1, 289–295.
    https://doi.org/10.1016/j.quageo.2007.01.004
.. [Ludwig1977]
    Ludwig, K.R., 1977. Effect of initial radioactive-daughter disequilibrium on
    U-Pb isotope apparent ages of young minerals. Journal of Research of the US
    Geological Survey 5, 663–667.
"""

import numpy as np

from . import cfg


exp = np.exp


#=====================================
# 234U age equaions
#=====================================
def ar48i(t, A48, l8, l4):
    """
    Initial 234U/238U activity ratio from present ratio.

    Notes
    ------
    Non-secular equilibrium equation is derived following approach of Ludwig
    (1977). This is an experimental feature only!

    """
    if cfg.secular_eq:
        return 1. + (A48 - 1.) * exp(l4 * t)
    else:
        return l4/(l4-l8) + (A48 - l4/(l4-l8)) *exp(-(l8-l4)*t)


def ar48(t, A48i, l8, l4):
    """
    234U/238U activity ratio from initial ratio and time.

    Notes
    ------
    Non-secular equilibrium equation is derived following approach of Ludwig
    (1977). This is an experimental feature only!

    """
    if cfg.secular_eq:
        return (A48i - 1.) * exp(-l4 * t) + 1.
    else:
        return l4/(l4-l8) + (A48i - l4/(l4-l8)) * exp((l8-l4)*t)


#=====================================
# Th230 equations
#=====================================

def Th230_age():
    pass


def Th230_age_uncert():
    pass


def ar08i(t, A48, A08, l8, l4, l0, init=True):
    """
    Initial [230Th/238U] as a function of present-day ratio and
    time. E.g. Cheng et al. (2000; GCA) and Hellstrom (2006; Q. Geochon.).

    Parameters
    ----------
    init : bool
        If True, [234U/238] is treated as an initial value, otherwise it's
        treated as a present-day value.

    Notes
    ------
    Non-secular equilibrium equation is derived following approach of Ludwig
    (1977). This is an experimental feature only!

    """
    if cfg.secular_eq:
        if init:
            A48i = A48
            A08i = 1. - (A48i-1.) * (l0/(l0-l4)) * (exp((l0-l4) * t) - 1) \
                    + (A08-1.) * exp(l0*t)
        else:
            A08i = 1. - (A48-1.) * (l0/(l0-l4)) * (exp(l0*t) - exp(l4*t)) \
                   + (A08-1.) * exp(l0 * t)
        return A08i
    else:
        a48i = ar48i(t, A48, l8, l4) if init else A48

        c1 = l0*l4 / ((l4-l8) * (l0-l8))
        c2 = l0*l4 / ((l8-l4) * (l0-l4))
        c3 = l0*l4 / ((l8-l0) * (l4-l0))

        A08i = (A08 - (c1 + c2*exp((l8-l4)*t) + c3*exp((l8-l0)*t)
                + a48i*(l0/(l0-l4)) * (exp((l8-l4)*t) - exp((l8-l0) * t)))) \
                * exp(-(l8-l0) * t)
        return A08i


def A08(t, A48, A08i, l8, l4, l0, init=True):
    """
    Present-day [230Th/238U] activity ratio as a function of initial ratio
    and time. Based on equations in, e.g., Cheng et al. (2000; GCA) and
    Hellstrom (2006; Q. Geochon.).

    Parameters
    ----------
    init : bool
        If True, [234U/238] is treated as an initial value, otherwise it's
        treated as a present-day value.

    Notes
    ------
    Non-secular equilibrium equation is derived following approach of Ludwig
    (1977). This is an experimental feature only!

    """
    if not cfg.secular_eq:
        raise ValueError('non-secular equations not yet implemented here')

    if init:
        A48i = A48
        A08 = 1. - exp(-l0*t) * (1. - A08i) + (A48i-1.)*(l0/(l0-l4)) \
              * (exp(-l4*t) - exp(-l0*t))
    else:
        A08 = 1. - exp(-l0*t) * (1. - A08i) + (A48-1.) * (l0/(l0-l4)) \
              * (1.-exp((l4-l0)*t))
    return A08


#=====================================
# U/Th isochron age equations
#=====================================

def Th230_isochron():
    pass



#==================================================
# Back-calculate initial activity ratio solutions
#===================================================

def Ai_solutions(t, A, init, DC):
    """
    Compute initial activity ratio solutions from present-day values and U-Pb age
    solution. In principle these values are computed iteratively along with the
    age solution, but practically it is more convenient to compute them after
    the fact.
    """
    a48i = None
    a08i = None
    if not init[0]:        # present [234U/238U]
        a48i = ar48i(t, A[0], DC[0], DC[1])
        if not init[1]:    # present [234U/238U] and [230Th/238U]
            a08i = ar08i(t, ar48i, A[1], DC[0], DC[1], DC[2])
    elif not init[1]:      # only present [230Th/238U]
        a08i = ar08i(t, A[0], A[1], DC[0], DC[1], DC[2], init=init[0])
    return a48i, a08i


"""
Weighted average algorithms for 1- and 2-dimensional data

References
----------
.. [Hayes2014]
    Hayes, K., 2014. "Finite-Sample Bias-Correction Factors for the Median
    Absolute Deviation." Communications in Statistics - Simulation and
    Computation 43, 2205–2212. https://doi.org/10.1080/03610918.2012.748913
.. [Lyons1988]
    Lyons, L., Gibaut, D., Clifford, P., 1988. How to combine correlated
    estimates of a single physical quantity. Nuclear Instruments and Methods in
    Physics Research Section A: Accelerators, Spectrometers, Detectors and
    Associated Equipment 270, 110–117.
    https://doi.org/10.1016/0168-9002(88)90018-6
.. [Maronna2019]
    Maronna, R.A., Martin, D., Yohai, V.J., Salibián-Barrera, M., 2019. Robust
    statistics. John Wiley and Sons, Chichester.
.. [McLean2011]
    McLean, N M, J F Bowring, and S A Bowring, 2011, An Algorithm for U-Pb Isotope
    Dilution Data Reduction and Uncertainty Propagation Geochemistry,
    Geophysics, Geosystems 12, no. 6.
    https://doi.org/10.1029/2010GC003478.
.. [Powell2020]
    Powell, R., Green, E.C.R., Marillo Sialer, E., Woodhead, J., 2020. Robust
    isochron calculation. Geochronology 2, 325–342.
    https://doi.org/10.5194/gchron-2-325-2020
.. [Powell1988]
    Powell, R., Holland, T.J.B., 1988. An internally consistent dataset with
    uncertainties and correlations: 3. Applications to geobarometry, worked
    examples and a computer program. J Metamorph Geol 6, 173–204.
    https://doi.org/10.1111/j.1525-1314.1988.tb00415.x

"""

import warnings
import numpy as np

from scipy.linalg import sqrtm


from . import cfg
from . import exceptions
from . import stats
from . import misc


# Simulated upper 95% confidence limit values on spine width for robust
# spine weighted average. Note, MAD has an odd/even number bias for small n
# (e.g. Hayes, 2014). Values greater than 30 are estimated based on curve fitting.
slim_even = {
    '4': 1.520,
    '6': 1.572,
    '8': 1.549,
    '10': 1.516,
    '12': 1.485,
    '14': 1.459,
    '16': 1.434,
    '18': 1.415,
    '20': 1.396,
    '22': 1.381,
    '24': 1.367,
    '26': 1.354,
    '28': 1.342,
    '30': 1.332,
}

slim_odd = {
    '3': 1.771,
    '5': 1.723,
    '7': 1.647,
    '9': 1.586,
    '11': 1.538,
    '13': 1.500,
    '15': 1.469,
    '17': 1.443,
    '19': 1.421,
    '21': 1.401,
    '23': 1.384,
    '25': 1.370,
    '27': 1.356,
    '29': 1.345,
}


#===========================================
# 1-D weighted average routines
#===========================================

def classical_wav(x, sx=None, cov=None, method='ca'):
    """
    Compute a classical weighted average accounting for assigned errors (and
    optionally error correlations).

    Partly emulates the behaviour of Isoplot (Ludwig, 2012). If one-sided MSWD
    confidence limit is above lower threshold (equivalent to a probability of
    fit below some upper threshold), uncertainty on weighted average is expanded
    according to data scatter (i.e., by sqrt(mswd)).

    Parameters
    ----------
    sx : np.ndarray, optional
        Analytical errors at the 1 sigma level.
    cov : np.ndarray, optional
        Error covariance matrix.

    Notes
    -----
    Does not yet include external error component if MSWD confidence limit is
    above an upper threshold (as in Isoplot MLE approach), although this
    feature may be added in future.

    """
    assert method in ('ca')                  # only one method at present
    assert (sx is not None) ^ (cov is not None)

    n = x.shape[0]
    assert n > 1
    if cov is None:
        if x.shape[0] != sx.shape[0]:
            raise ValueError('shape mismatch between inputs sx and x')
        xbar, sxbar, mswd, r = wav(x, sx)
    else:
        if cov.shape != (n, n):
            raise ValueError('shape mismatch between inputs cov and x')
        xbar, sxbar, mswd, r = wav_cor(x, cov)

    # get pr fit:
    df = n - 1
    p = stats.pr_fit(df, mswd)

    # get mswd_regression_thresholds:
    mswd_threshold_1 = stats.mswd_conf_limits(df, p=cfg.mswd_ci_thresholds[0])
    mswd_threshold_2 = stats.mswd_conf_limits(df, p=cfg.mswd_ci_thresholds[1])

    # equivalent p thresholds -
    p_threshold_1 = 1. - cfg.mswd_ci_thresholds[0]
    p_threshold_2 = 1. - cfg.mswd_ci_thresholds[1]

    # If p is below limit, then expand errors to account for excess scatter.
    excess_scatter = True if mswd > mswd_threshold_1 else False
    t_crit = stats.t_critical(df) if excess_scatter else 1.96
    scatter_mult = np.sqrt(mswd) if excess_scatter else 1.

    results = {
        'type': 'classical',
        'n': n,
        'cov': True if cov is not None else False,
        'ave': xbar,
        'ave_1s': sxbar,
        'ave_95pm': t_crit * scatter_mult * sxbar,
        'mswd':  mswd,
        'mswd_regression_thresholds': [mswd_threshold_1, mswd_threshold_2],
        'p': p,
        'p_thresholds': [p_threshold_1, p_threshold_2],
        't_crit': t_crit,
        'excess_scatter': excess_scatter,
        'wtd_residuals': r
    }
    return results


def robust_wav(x, sx=None, cov=None, method='ra'):
    """
    Compute robust weighted average using the spine algorithm.
    """
    assert method in ('rs', 'ra')
    assert (sx is not None) ^ (cov is not None)

    n = x.size
    assert n > 1
    if n < 3:
        raise ValueError('n must be greater than or equal to 3')

    model = 'spine'

    if cov is None:
        if x.shape[0] != sx.shape[0]:
            raise ValueError('shape mismatch between inputs cov and x')
        xbar, sxbar, s, r = spine_wav(x, sx, h=cfg.h)
    else:
        if cov.shape != (n, n):
            raise ValueError('shape mismatch between inputs cov and x')
        xbar, sxbar, s, r = spine_wav_cor(x, cov, h=cfg.h)
    slim = s95lim(n)

    # Check for excess scatter (note definition of excess scatter here
    # different from classical stats weighted average):
    if s > slim:
        excess_scatter = True
    else:
        excess_scatter = False

    results = {
        'type': 'robust',
        'model': model,
        'n': n,
        'cov': True if cov is not None else False,
        'ave': xbar,
        'ave_1s': sxbar,
        'ave_95pm': 1.96 * sxbar,
        's': s,
        's_upper_95ci': slim,
        'excess_scatter': excess_scatter,
        'wtd_residuals': r,
    }
    return results


#===========================================
# 1-D classical weighted average functions
#===========================================

def wav(x, sx):
    """
    Basic clossical stats weighted average without accounting for error
    correlations.
    """
    w = 1. / sx ** 2
    xb = np.sum(w * x) / np.sum(w)                # weighted average x
    sxb = 1. / np.sqrt(np.sum(w))                 # std error on weighted average

    e = x - xb
    r = e / sx                                    # weighted residuals
    s = np.sum(r ** 2)
    df = x.shape[0] - 1
    mswd = s / df

    return xb, sxb, mswd, r


def wavx():
    """
    Classical weighted average with excess scatter parameterised as a uniform
    Gaussian source of error applied equally to all datapoints.
    """
    return


def wav_cor(x, V):
    """
    Compute classical stats weighted mean accounting for error correlations
    amongst individual measurements.

    Based on equations in, e.g., Powell and Holland (1988), Lyons et al
    (1988), and McLean (2011).

    Notes
    ------
    See cautionary note in Lyons et al. (1988):

        "One point to beware of in practice is
        the situation in which the individual weights become numerically very
        large." ... this happens when [errors are very close to each other and
        highly correlated]. If we have slightly mis-estimated the elements of
        the error matrix, or if our measurements are slightly biassed, the effect
        of the large weights with different signs can be to drive our solution
        far away from the correct value."

    Parameters
    --------
    x : np.ndarray
        Data points to be averaged (should be a 1-dimensional array)
    V : np.ndarray
        covariance matrix (should be a 2-dimensional array)
    """
    if not misc.pos_def(V):
        raise ValueError('V must be a positive definite matrix')

    n = x.shape[0]
    avx = np.mean(x)
    x = x / avx
    V = V / avx ** 2

    x = x.reshape(n, 1)                # reshape to column vector
    ones = np.ones(x.shape)
    V_inv = np.linalg.inv(V)

    denom = ones.T @ V_inv @ ones
    alpha = V_inv @ ones / denom
    xb = float(alpha.T @ x)                     # wtd. average
    sxb = float(np.sqrt(1. / denom))            # standard error on wtd. average

    # goodness of fit
    e = x - xb
    s = e.T @ np.linalg.inv(V) @ e
    df = n - 1.
    mswd = float(s / df)

    # weighted residuals
    V_inv2 = sqrtm(V_inv)                           # V^(-1.2) - analytical weights
    r = V_inv2 @ e
    assert np.isclose(sum(r ** 2) / df, mswd)       # sanity check

    x *= avx
    xb *= avx
    sxb *= avx

    return xb, sxb, mswd, r.flatten()


#==============================
# 1-D robust weighted average
#==============================

def s95lim(n):
    """
    Upper 95% confidence limit on spine width (s) for weighted averages.
    Values are derived via simulation of Gaussian distributed datasets.
    """
    if n < 4:
        return np.nan

    elif n <= 30:                           # lookup value in table
        if n % 2 == 0:
            return slim_even[str(n)]
        else:
            return slim_odd[str(n)]
    else:                                   # use curve fit estimation
        if n > 150:
            warnings.warn('95% confidence limit on s may be underestimated for '
                          'n > 150')
        if n % 2 == 0:
            return 1.52825321 - 0.07689812 * np.log(-17.63329986 + n)
        else:
            return 1.53489602 - 0.07829274 * np.log(-18.25797641 + n)


def spine_wav(x, sx, xb0=None, maxiter=50, h=1.4):
    """
    Spine robust weighted average **without** accounting for error covariance.
    """
    avx = np.mean(x)
    x = x / avx
    sx = sx / avx

    # solve for xbar using iterative re-weighting (e.g. sect. 2.8 of Maronna)
    xb0 = np.median(x) if xb0 is None else xb0
    i = 0

    while i < maxiter:
        i += 1
        e = x - xb0
        r = e / sx                              # weighted residuals
        wh = [1 if abs(rk) < h else np.sqrt(h / abs(rk))
              for rk in r.flatten()]
        w = wh / sx ** 2                        # see e.g. B12 of Powell (2020)
        xb = sum(w * x) / sum(w)

        if np.isclose(xb, xb0, atol=0, rtol=1e-10):

            dpsi = [1 if abs(rk) < h else 0 for rk in r]
            var = 1. / sum(dpsi / sx ** 2)
            sxb = np.sqrt(var)  # standard error on wtd average
            s = stats.nmad(r)  # "spine width"

            xb *= avx
            sxb *= avx

            return xb, sxb, s, r

        xb0 = xb

    raise exceptions.ConvergenceError('Spine weighted average routine failed to converge '
            'after maximum number of iterations')


def spine_wav_cor(x, V, xb0=None, maxiter=50, h=1.4, atol=1e-08, rtol=1e-08):
    """
    Spine robust weighted avareage. Equivalent to a 1-dimensional version
    of the spine linear regression algorithm of Powell et al., (2020), accounting
    for error covariances. Reduces to classical stats solution for "well-behaved"
    datasets.
    """
    n = x.shape[0]

    if not misc.pos_def(V):
        raise ValueError("V must be a positive definite matrix")

    avx = np.mean(x)
    x = x / avx
    V = V / avx ** 2

    x = x.reshape(n, 1)                         # reshape to column vector
    ones = np.ones(x.shape)
    V_inv = np.linalg.inv(V)
    V_inv2 = sqrtm(V_inv)               # V^(-1/2) - analytical weights
                                        # without sorting eigenv(alues)ectors

    # Double check that eigenvalues and vectors are in correct order:
    assert np.allclose(V_inv2 @ V_inv2, V_inv)

    xb0 = np.median(x) if xb0 is None else xb0
    i = 0

    # solve for x-bar using iterative re-weighting (e.g. sect. 2.8 of Maronna)
    while i < maxiter:
        i += 1
        e = x - xb0
        r = V_inv2 @ e                              # uncorrelated weighted residuals
        wh = [1 if abs(rk) < h else np.sqrt(h / abs(rk))
              for rk in r.flatten()]
        w = V_inv @ np.diag(wh)                     # e.g. B12 of Powell (2020)
        xb = ones.T @ w @ x / (ones.T @ w @ ones)   # e.g. B13 of Powell (2020)

        if np.isclose(xb, xb0, atol=atol, rtol=rtol):

            dpsi = [1 if abs(rk) < h else 0 for rk in r]
            var = 1. / (ones.T @ V_inv @ np.diag(dpsi) @ ones)
            sxb = np.sqrt(var)                     # standard error on wtd average
            s = stats.nmad(r)                      # spine width

            xb = float(xb) * avx
            sxb = float(sxb) * avx
            r = r.reshape(n, )

            return xb, sxb, s, r

        xb0 = xb

    raise exceptions.ConvergenceError('Spine weighted average routine failed '
            'to converge after maximum number of iterations')


#=================================
# 2-D weighted average
#=================================

# def wav_2d():
#     """Two-dimensional weighted average based on classical stats.
#     """
#     return


# def spine_wav_2d():
#     """
#     Would this be useful?
#     It would require a robust version of MSWD equivalence, then perhaps if
#     data points are deemed equivalent, the classical stats approach could still
#     be used to assess MSWD concordance and compute age uncertainty?
#     """
#     return



"""
A Python library for geochronology.

"""

from . import uxpb
from .exceptions import ConvergenceError
from . import cfg
from . import concordia
from . import data
from . import isochron
from . import misc
from . import monte_carlo
from . import plotting
from . import regression
from . import stats
from . import transform
from . import upb
from . import useries
from . import wtd_average






"""
Pb/U equations assuming disequilibrium in a single nuclide only

References
----------
.. [Guillong2014]
    Guillong, M., von Quadt, A., Sakata, S., Peytcheva, I., Bachmann, O., 2014.
    LA-ICP-MS Pb-U dating of young zircons from the Kos–Nisyros volcanic centre,
    SE aegean arc. Journal of Analytical Atomic Spectrometry 29, 963–970.
    https://doi.org/10.1039/C4JA00009A
.. [Sakata2017]
    Sakata, S., Hirakawa, S., Iwano, H., Danhara, T., Guillong, M., Hirata, T.,
    2017. A new approach for constraining the magnitude of initial
    disequilibrium in Quaternary zircons by coupled uranium and thorium decay
    series dating. Quaternary Geochronology 38, 1–12.
    https://doi.org/10.1016/j.quageo.2016.11.002

"""

import numpy as np


def f(t, fThU, l8, l0):
    """
    206Pb/238U atomic ratio as a function of time and actvitiy
    ratios using equation from Guillong et al. (2014).
    """
    return np.exp(l8 * t) - 1. + l8 / l0 * (fThU - 1) \
           * (1 - np.exp(-l0 * t)) * np.exp(l8 * t)


def g(t, fPaU, l5, l1):
    """
    207Pb/235U atomic ratio as a function of time and actvitiy
    ratios using equation from Sakata et al. (2018).
    """
    return (np.exp(l5 * t) - 1.) + l5 / l1 * (fPaU - 1) \
           * (1 - np.exp(-l1 * t)) * np.exp(l5 * t)
